import numpy
from enum import Enum
import math
import logging

from data_specification.enums.data_type import DataType
from pacman.executor.injection_decorator import inject_items, \
    supports_injection
from pacman.model.graphs.common.slice import Slice
from pacman.model.graphs.machine import MachineVertex
from pacman.model.resources import ResourceContainer, SDRAMResource, \
    CPUCyclesPerTickResource, DTCMResource

from spinn_front_end_common.abstract_models.\
    abstract_generates_data_specification import \
    AbstractGeneratesDataSpecification
from spinn_front_end_common.abstract_models.\
    abstract_has_associated_binary import \
    AbstractHasAssociatedBinary
from spinn_front_end_common.interface.buffer_management import \
    recording_utilities
from spinn_front_end_common.interface.buffer_management.buffer_models.\
    abstract_receive_buffers_to_host import \
    AbstractReceiveBuffersToHost
from spinn_front_end_common.utilities import globals_variables
from spinn_front_end_common.utilities.utility_objs.\
    executable_start_type import \
    ExecutableStartType
from spinn_front_end_common.utilities import constants
from spinn_front_end_common.utilities import helpful_functions
from spinn_front_end_common.interface.simulation import simulation_utilities

from spinn_utilities.overrides import overrides

logger = logging.getLogger(__name__)


@supports_injection
class ChipPowerMonitorMachineVertex(
        MachineVertex, AbstractHasAssociatedBinary,
        AbstractGeneratesDataSpecification, AbstractReceiveBuffersToHost):
    """ machine vertex for c code representing functionality to record /
            idle times in a machine graph
    """

    # data regions
    CHIP_POWER_MONITOR_REGIONS = Enum(
        value="CHIP_POWER_MONITOR_REGIONS",
        names=[('SYSTEM', 0),
               ('CONFIG', 1),
               ('RECORDING', 2)])

    # default magic numbers
    DEFAULT_MALLOCS_USED = 3
    CONFIG_SIZE_IN_BYTES = 8
    RECORDING_SIZE_PER_ENTRY = 18 * 4
    SAMPLE_RECORDING_REGION = 0
    MAX_CORES_PER_CHIP = 18
    MAX_BUFFER_SIZE = 1048576

    def __init__(
            self, label, constraints, n_samples_per_recording,
            sampling_frequency):
        """ constructor for idle power monitor c code for machine graphs

        :param label: vertex label
        :param constraints: constraints on this vertex
        :param n_samples_per_recording: how may samples between recording entry
        :type n_samples_per_recording: int
        :param sampling_frequency: how often to sample
        :type sampling_frequency: microseconds
        """
        MachineVertex.__init__(self, label=label, constraints=constraints)
        self._n_samples_per_recording = n_samples_per_recording
        self._sampling_frequency = sampling_frequency

    @property
    def sampling_frequency(self):
        return self._sampling_frequency

    @property
    def n_samples_per_recording(self):
        return self._n_samples_per_recording

    @property
    @inject_items({"machine_time_step": "MachineTimeStep",
                   "n_machine_time_steps": "TotalMachineTimeSteps",
                   "time_scale_factor": "TimeScaleFactor"})
    @overrides(MachineVertex.resources_required,
               additional_arguments={
                   'machine_time_step', 'n_machine_time_steps',
                   'time_scale_factor'})
    def resources_required(
            self, n_machine_time_steps, machine_time_step, time_scale_factor):
        return self.get_resources(
            n_machine_time_steps, machine_time_step, time_scale_factor,
            self._n_samples_per_recording, self._sampling_frequency)

    @staticmethod
    def get_resources(
            n_machine_time_steps, time_step, time_scale_factor,
            n_samples_per_recording, sampling_frequency):
        """ get resources used by this vertex

        :return:Resource container
        """
        # get config
        config = globals_variables.get_simulator().config

        # get recording params
        minimum_buffer_sdram = config.getint(
            "Buffers", "minimum_buffer_sdram")
        using_auto_pause_and_resume = config.getboolean(
            "Buffers", "use_auto_pause_and_resume")
        receive_buffer_host = config.get("Buffers", "receive_buffer_host")
        receive_buffer_port = helpful_functions.read_config_int(
            config, "Buffers", "receive_buffer_port")

        # figure recording size for max run
        if not using_auto_pause_and_resume and n_machine_time_steps is None:
            raise Exception(
                "You cannot use the chip power montiors without auto pause "
                "and resume and not allocating a n_machine_time_steps")

        # figure max buffer size
        max_buffer_size = 0
        if config.getboolean("Buffers", "enable_buffered_recording"):
            max_buffer_size = config.getint(
                "Buffers", "chip_power_monitor_buffer")

        maximum_sdram_for_buffering = [max_buffer_size]

        n_recording_entries = (math.ceil(
            (sampling_frequency / (time_step * time_scale_factor))) /
            n_samples_per_recording)

        recording_size = (
            ChipPowerMonitorMachineVertex.RECORDING_SIZE_PER_ENTRY *
            n_recording_entries)

        container = ResourceContainer(
            sdram=SDRAMResource(
                ChipPowerMonitorMachineVertex.sdram_calculation()),
            cpu_cycles=CPUCyclesPerTickResource(100),
            dtcm=DTCMResource(100))
        recording_sizes = recording_utilities.get_recording_region_sizes(
            [int(recording_size)], n_machine_time_steps, minimum_buffer_sdram,
            maximum_sdram_for_buffering, using_auto_pause_and_resume)
        container.extend(recording_utilities.get_recording_resources(
            recording_sizes, receive_buffer_host, receive_buffer_port))
        return container

    @staticmethod
    def sdram_calculation():
        """ calculates the sdram requirements of the vertex

        :return: int
        """
        return constants.SYSTEM_BYTES_REQUIREMENT + \
            ChipPowerMonitorMachineVertex.CONFIG_SIZE_IN_BYTES + \
            ChipPowerMonitorMachineVertex.DEFAULT_MALLOCS_USED * \
            constants.SARK_PER_MALLOC_SDRAM_USAGE

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self):
        return ChipPowerMonitorMachineVertex.binary_file_name()

    @staticmethod
    def binary_file_name():
        """ returns the string binary file name

        :return: basestring
        """
        return "chip_power_monitor.aplx"

    @inject_items({"time_scale_factor": "TimeScaleFactor",
                   "machine_time_step": "MachineTimeStep",
                   "n_machine_time_steps": "TotalMachineTimeSteps",
                   "ip_tags": "MemoryIpTags"})
    @overrides(AbstractGeneratesDataSpecification.generate_data_specification,
               additional_arguments={
                   "machine_time_step", "time_scale_factor",
                   "n_machine_time_steps", "ip_tags"})
    def generate_data_specification(
            self, spec, placement, machine_time_step, time_scale_factor,
            n_machine_time_steps, ip_tags):
        self._generate_data_specification(
            spec, machine_time_step, time_scale_factor, n_machine_time_steps,
            ip_tags)

    def _generate_data_specification(
            self, spec, machine_time_step, time_scale_factor,
            n_machine_time_steps, ip_tags):
        """ this is used to support application vertex calling this directly

        :param spec: data spec
        :param machine_time_step: machine time step
        :param time_scale_factor: time scale factor
        :param n_machine_time_steps: n_machine time steps
        :param ip_tags: iptags
        :rtype: None
        """
        spec.comment("\n*** Spec for ChipPowerMonitor Instance ***\n\n")

        # Construct the data images needed for the Neuron:
        self._reserve_memory_regions(spec)
        self._write_setup_info(
            spec, machine_time_step, time_scale_factor, n_machine_time_steps,
            Slice(0, 1), ip_tags)
        self._write_configuration_region(spec)

        # End-of-Spec:
        spec.end_specification()

    def _write_configuration_region(self, spec):
        """ writes the data needed by the c code to configure itself

        :param spec: spec object
        :rtype: None
        """
        spec.switch_write_focus(
            region=ChipPowerMonitorMachineVertex.CHIP_POWER_MONITOR_REGIONS.
            CONFIG.value)
        spec.write_value(self._n_samples_per_recording,
                         data_type=DataType.UINT32)
        spec.write_value(self._sampling_frequency, data_type=DataType.UINT32)

    def _write_setup_info(
            self, spec, machine_time_step, time_scale_factor,
            n_machine_time_steps, vertex_slice, ip_tags):
        """ writes the system data as required

        :param spec: the dsg spec writer
        :param machine_time_step: the machine time step
        :param time_scale_factor: the time scale factor
        :rtype: None
        """
        spec.switch_write_focus(
            region=(
                ChipPowerMonitorMachineVertex.CHIP_POWER_MONITOR_REGIONS.
                SYSTEM.value))
        spec.write_array(simulation_utilities.get_simulation_header_array(
            self.get_binary_file_name(), machine_time_step, time_scale_factor))

        spec.switch_write_focus(
            ChipPowerMonitorMachineVertex.CHIP_POWER_MONITOR_REGIONS.
            RECORDING.value)
        recorded_region_sizes = recording_utilities.get_recorded_region_sizes(
            n_machine_time_steps,
            [self._deduce_sdram_requirements_per_timer_tick(
                machine_time_step, time_scale_factor)],
            [self.MAX_BUFFER_SIZE])
        spec.write_array(recording_utilities.get_recording_header_array(
            recorded_region_sizes,
            globals_variables.get_simulator().config.getint(
                "Buffers", "time_between_requests"),
            None, ip_tags))

    def _reserve_memory_regions(self, spec):
        """ reserve the dsg memory regions as required

        :param spec: the dsg specification to reserve in
        :rtype: None
        """
        spec.comment("\nReserving memory space for data regions:\n\n")

        # Reserve memory:
        spec.reserve_memory_region(
            region=(
                ChipPowerMonitorMachineVertex.CHIP_POWER_MONITOR_REGIONS.
                SYSTEM.value),
            size=constants.SYSTEM_BYTES_REQUIREMENT,
            label='system')
        spec.reserve_memory_region(
            region=(
                ChipPowerMonitorMachineVertex.CHIP_POWER_MONITOR_REGIONS.
                CONFIG.value),
            size=self.CONFIG_SIZE_IN_BYTES, label='config')
        spec.reserve_memory_region(
            region=(ChipPowerMonitorMachineVertex.
                    CHIP_POWER_MONITOR_REGIONS.RECORDING.value),
            size=recording_utilities.get_recording_header_size(1),
            label="Recording")

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self):
        return self.binary_start_type()

    @staticmethod
    def binary_start_type():
        """ static method to allow app verts to use this

        :return: starttype enum
        """
        return ExecutableStartType.USES_SIMULATION_INTERFACE

    @overrides(AbstractReceiveBuffersToHost.get_recording_region_base_address)
    def get_recording_region_base_address(self, txrx, placement):
        return helpful_functions.locate_memory_region_for_placement(
            placement,
            ChipPowerMonitorMachineVertex.CHIP_POWER_MONITOR_REGIONS.
            RECORDING.value, txrx)

    @overrides(AbstractReceiveBuffersToHost.get_recorded_region_ids)
    def get_recorded_region_ids(self):
        return [0]

    @inject_items({"time_scale_factor": "TimeScaleFactor"})
    @overrides(AbstractReceiveBuffersToHost.get_n_timesteps_in_buffer_space,
               additional_arguments={"time_scale_factor"})
    def get_n_timesteps_in_buffer_space(
            self, buffer_space, machine_time_step, time_scale_factor):
        return recording_utilities.get_n_timesteps_in_buffer_space(
            buffer_space,
            [self._deduce_sdram_requirements_per_timer_tick(
                machine_time_step, time_scale_factor)])

    @inject_items({"machine_time_step": "MachineTimeStep",
                   "n_machine_time_steps": "TotalMachineTimeSteps",
                   "time_scale_factor": "TimeScaleFactor"})
    @overrides(AbstractReceiveBuffersToHost.get_minimum_buffer_sdram_usage,
               additional_arguments={
                   'machine_time_step', 'n_machine_time_steps',
                   'time_scale_factor'})
    def get_minimum_buffer_sdram_usage(
            self, n_machine_time_steps, machine_time_step, time_scale_factor):
        return recording_utilities.get_minimum_buffer_sdram(
            [self._deduce_sdram_requirements_per_timer_tick(
                machine_time_step, time_scale_factor)],
            n_machine_time_steps,
            globals_variables.get_simulator().config.getint(
                "Buffers", "minimum_buffer_sdram"))[0]

    def _deduce_sdram_requirements_per_timer_tick(
            self, machine_time_step, time_scale_factor):
        """ deduce sdram usage per timer tick

        :param machine_time_step: the machine time step
        :param time_scale_factor: the time scale factor
        :return: the sdram usage
        """
        timer_tick_in_micro_seconds = machine_time_step * time_scale_factor
        recording_time = \
            self._sampling_frequency * self._n_samples_per_recording
        n_entries = math.floor(timer_tick_in_micro_seconds / recording_time)
        return math.ceil(
            n_entries * ChipPowerMonitorMachineVertex.RECORDING_SIZE_PER_ENTRY)

    def get_recorded_data(self, placement, buffer_manager):
        """ get data from sdram given placement and buffer manager

        :param placement: the location on machien to get data from
        :param buffer_manager: the buffer manager that might have data
        :return: results
        :rtype: numpy array with 1 dimension
        """
        # for buffering output info is taken form the buffer manager
        samples, data_missing = \
            buffer_manager.get_data_for_vertex(
                placement,
                ChipPowerMonitorMachineVertex.SAMPLE_RECORDING_REGION)
        if data_missing:
            logger.warn(
                "Chip Power monitor has lost data on chip({}, {})"
                .format(placement.x, placement.y))

        # get raw data as a byte array
        record_raw = samples.read_all()

        results = (
            numpy.frombuffer(record_raw, dtype="uint32").reshape(-1, 18) /
            self.n_samples_per_recording)
        return results



# general imports
import logging
from six import add_metaclass

# spinn front end common imports
from .abstract_sends_buffers_from_host import AbstractSendsBuffersFromHost

from spinn_utilities.abstract_base import AbstractBase, abstractmethod, \
    abstractproperty

logger = logging.getLogger(__name__)


@add_metaclass(AbstractBase)
class SendsBuffersFromHostPreBufferedImpl(AbstractSendsBuffersFromHost):
    """ Implementation of the AbstractSendsBuffersFromHost\
        which uses an existing set of buffers for the details
    """

    __slots__ = ()

    @abstractproperty
    def send_buffers(self):
        pass

    @send_buffers.setter
    @abstractmethod
    def send_buffers(self, value):
        pass

    def buffering_input(self):
        return self.send_buffers is not None

    def get_regions(self):
        """ Return the regions which has buffers to send
        """
        return self.send_buffers.keys()

    def get_max_buffer_size_possible(self, region):
        """ Return the max possible size of a buffered region

        :param region: the region to find the max possible size of
        :return: the max possible size of the buffered region
        """
        return self.send_buffers[region].max_buffer_size_possible

    def get_region_buffer_size(self, region):
        """ Return the size of a given regions buffer

        :param region: the region to find the size of
        :return: the size of the buffer
        """
        return self.send_buffers[region].buffer_size

    def is_next_timestamp(self, region):
        """ Check if there are more time stamps which need transmitting

        :param region: the region to check
        :return: boolean
        """
        return self.send_buffers[region].is_next_timestamp

    def get_next_timestamp(self, region):
        """ Return the next time stamp available in the buffered region

        :param region: the region id which is being asked
        :return: the next time stamp
        """
        return self.send_buffers[region].next_timestamp

    def is_next_key(self, region, timestamp):
        """ Check if there is more keys to transmit for a given region in a\
            given timestamp

        :param region: the region id to check
        :param timestamp:  the timestamp to check
        :return: bool
        """
        return self.send_buffers[region].is_next_key(timestamp)

    def get_next_key(self, region):
        """ Get the next key for a given region

        :param region: the region to get the next key from
        """
        return self.send_buffers[region].next_key

    def is_empty(self, region):
        """ Check if a region is empty

        :param region: the region id to check
        :return: bool
        """
        return len(self.send_buffers[region].timestamps) == 0

    def rewind(self, region):
        """ Rewinds the internal buffer in preparation of re-sending\
            the spikes

        :param region: The region to rewind
        :type region: int
        """
        self.send_buffers[region].rewind()



from spinn_utilities.progress_bar import ProgressBar

# data spec imports
from data_specification import DataSpecificationExecutor, constants
from data_specification.exceptions import DataSpecificationException

# spinn_storage_handlers import
from spinn_storage_handlers import FileDataReader

import logging
import struct
import numpy

logger = logging.getLogger(__name__)


class HostExecuteDataSpecification(object):
    """ Executes the host based data specification
    """

    __slots__ = []

    def __call__(self, transceiver, machine, app_id, dsg_targets):
        """

        :param machine: the python representation of the spinnaker machine
        :param transceiver: the spinnman instance
        :param app_id: the application ID of the simulation
        :param dsg_targets: map of placement to file path

        :return: map of placement and dsg data, and loaded data flag.
        """
        processor_to_app_data_base_address = dict()

        # create a progress bar for end users
        progress = ProgressBar(
            dsg_targets, "Executing data specifications and loading data")

        for ((x, y, p), data_spec_file_path) in progress.over(
                dsg_targets.iteritems()):

            # build specification reader
            data_spec_file_path = dsg_targets[x, y, p]
            data_spec_reader = FileDataReader(data_spec_file_path)

            # maximum available memory
            # however system updates the memory available
            # independently, so the check on the space available actually
            # happens when memory is allocated
            chip = machine.get_chip_at(x, y)
            memory_available = chip.sdram.size

            # generate data spec executor
            executor = DataSpecificationExecutor(
                data_spec_reader, memory_available)

            # run data spec executor
            try:
                # bytes_used_by_spec, bytes_written_by_spec = \
                executor.execute()
            except DataSpecificationException as e:
                logger.error(
                    "Error executing data specification for {}, {}, {}".format(
                        x, y, p))
                raise e

            bytes_used_by_spec = executor.get_constructed_data_size()

            # allocate memory where the app data is going to be written
            # this raises an exception in case there is not enough
            # SDRAM to allocate
            start_address = transceiver.malloc_sdram(
                x, y, bytes_used_by_spec, app_id)

            # Write the header and pointer table and load it
            header = executor.get_header()
            pointer_table = executor.get_pointer_table(start_address)
            data_to_write = numpy.concatenate(
                (header, pointer_table)).tostring()
            transceiver.write_memory(x, y, start_address, data_to_write)
            bytes_written_by_spec = len(data_to_write)

            # Write each region
            for region_id in range(constants.MAX_MEM_REGIONS):
                region = executor.get_region(region_id)
                if region is not None:

                    max_pointer = region.max_write_pointer
                    if not region.unfilled and max_pointer > 0:

                        # Get the data up to what has been written
                        data = region.region_data[:max_pointer]

                        # Write the data to the position
                        position = pointer_table[region_id]
                        transceiver.write_memory(x, y, position, data)
                        bytes_written_by_spec += len(data)

            # set user 0 register appropriately to the application data
            user_0_address = \
                transceiver.get_user_0_register_address_from_core(x, y, p)
            start_address_encoded = \
                buffer(struct.pack("<I", start_address))
            transceiver.write_memory(
                x, y, user_0_address, start_address_encoded)

            # write information for the memory map report
            processor_to_app_data_base_address[x, y, p] = {
                'start_address': start_address,
                'memory_used': bytes_used_by_spec,
                'memory_written': bytes_written_by_spec}

        return processor_to_app_data_base_address, True



from .application_finisher import ApplicationFinisher
from .application_runner import ApplicationRunner
from .buffer_extractor import BufferExtractor
from .buffer_manager_creator import BufferManagerCreator
from .chip_iobuf_clearer import ChipIOBufClearer
from .chip_iobuf_extractor import ChipIOBufExtractor
from .chip_provenance_updater import ChipProvenanceUpdater
from .chip_runtime_updater import ChipRuntimeUpdater
from .database_interface import DatabaseInterface
from .dsg_region_reloader import DSGRegionReloader
from .edge_to_n_keys_mapper import EdgeToNKeysMapper
from .graph_binary_gatherer import GraphBinaryGatherer
from .graph_data_specification_writer \
    import GraphDataSpecificationWriter
from .graph_measurer import GraphMeasurer
from .graph_provenance_gatherer import GraphProvenanceGatherer
from .hbp_allocator import HBPAllocator
from .hbp_max_machine_generator import HBPMaxMachineGenerator
from .host_execute_data_specification \
    import HostExecuteDataSpecification
from .insert_edges_to_live_packet_gatherers \
    import InsertEdgesToLivePacketGatherers
from .insert_live_packet_gatherers_to_graphs \
    import InsertLivePacketGatherersToGraphs
from .load_executable_images import LoadExecutableImages
from .locate_executable_start_type import LocateExecutableStartType
from .machine_execute_data_specification \
    import MachineExecuteDataSpecification
from .machine_generator import MachineGenerator
from .notification_protocol import NotificationProtocol
from .placements_provenance_gatherer \
    import PlacementsProvenanceGatherer
from .pre_allocate_resources_for_live_packet_gatherers \
    import PreAllocateResourcesForLivePacketGatherers
from .provenance_json_writer import ProvenanceJSONWriter
from .provenance_xml_writer import ProvenanceXMLWriter
from .router_provenance_gatherer import RouterProvenanceGatherer
from .routing_setup import RoutingSetup
from .routing_table_loader import RoutingTableLoader
from .spalloc_allocator import SpallocAllocator
from .spalloc_max_machine_generator \
    import SpallocMaxMachineGenerator
from .tags_loader import TagsLoader
from .tdma_agenda_builder import TDMAAgendaBuilder
from .virtual_machine_generator import VirtualMachineGenerator

__all__ = [
    "ApplicationFinisher",
    "ApplicationRunner", "BufferExtractor",
    "BufferManagerCreator", "ChipIOBufClearer",
    "ChipIOBufExtractor", "ChipProvenanceUpdater",
    "ChipRuntimeUpdater", "DatabaseInterface",
    "DSGRegionReloader", "EdgeToNKeysMapper",
    "GraphBinaryGatherer",
    "GraphDataSpecificationWriter",
    "GraphMeasurer", "GraphProvenanceGatherer",
    "HBPAllocator", "HBPMaxMachineGenerator",
    "HostExecuteDataSpecification",
    "InsertEdgesToLivePacketGatherers",
    "InsertLivePacketGatherersToGraphs",
    "LoadExecutableImages",
    "LocateExecutableStartType",
    "MachineExecuteDataSpecification",
    "MachineGenerator", "NotificationProtocol",
    "PlacementsProvenanceGatherer",
    "PreAllocateResourcesForLivePacketGatherers",
    "ProvenanceJSONWriter", "ProvenanceXMLWriter",
    "RouterProvenanceGatherer", "RoutingSetup",
    "RoutingTableLoader", "SpallocAllocator",
    "SpallocMaxMachineGenerator", "TagsLoader",
    "TDMAAgendaBuilder",
    "VirtualMachineGenerator"]



import logging
import requests
import sys
from threading import Thread

from spinn_front_end_common.abstract_models \
    import AbstractMachineAllocationController

logger = logging.getLogger(__name__)


class _HBPJobController(Thread, AbstractMachineAllocationController):
    __slots__ = [
        # thread flag to allow it to be killed when the main thread dies
        "daemon",

        # the URLs to call the HBP system
        "_extend_lease_url",
        "_check_lease_url",

        # boolean flag for telling this thread when the system has ended
        "_exited"
    ]

    _WAIT_TIME_MS = 10000

    def __init__(self, url):
        Thread.__init__(self, name="HBPJobController")
        self.daemon = True
        self._extend_lease_url = "{}/extendLease".format(url)
        self._check_lease_url = "{}/checkLease".format(url)
        self._exited = False

        # Lower the level of requests to WARNING to avoid extra messages
        logging.getLogger("requests").setLevel(logging.WARNING)

    def extend_allocation(self, new_total_run_time):
        requests.get(self._extend_lease_url, params={
            "runTime": new_total_run_time})

    def _check_lease(self, wait_time):
        return requests.get(self._check_lease_url, params={
            "waitTime": wait_time}).json()

    def close(self):
        self._exited = True

    @property
    def power(self):
        # TODO NEEDS FIXING
        return True

    def set_power(self, power):
        # TODO NEEDS FIXING
        pass

    def where_is_machine(self, chip_x, chip_y):
        # TODO NEEDS FIXING
        pass

    def run(self):
        job_allocated = True
        while job_allocated and not self._exited:
            job_allocated = self._check_lease(self._WAIT_TIME_MS)["allocated"]

        if not self._exited:
            logger.error(
                "The allocated machine has been released before the end of"
                " the script - this script will now exit")
            sys.exit(1)


class HBPAllocator(object):
    """ Request a machine from the HBP remote access server that will fit\
        a number of chips
    """

    def __call__(
            self, hbp_server_url, total_run_time, n_chips):
        """

        :param hbp_server_url: The URL of the HBP server from which to get\
                    the machine
        :param total_run_time: The total run time to request
        :param n_chips: The number of chips required
        """

        url = hbp_server_url
        if url.endswith("/"):
            url = url[:-1]

        machine = self._get_machine(url, n_chips, total_run_time)
        hbp_job_controller = _HBPJobController(url)
        hbp_job_controller.start()

        bmp_details = None
        if "bmp_details" in machine:
            bmp_details = machine["bmpDetails"]

        return (
            machine["machineName"], int(machine["version"]), None, None,
            bmp_details, False, False, None, None, None,
            hbp_job_controller, None
        )

    def _get_machine(self, url, n_chips, total_run_time):
        get_machine_request = requests.get(
            url, params={"nChips": n_chips, "runTime": total_run_time})
        return get_machine_request.json()



from pacman.executor.injection_decorator import inject_items
from spinn_front_end_common.abstract_models \
    import AbstractProvidesKeyToAtomMapping
from pacman.model.decorators import overrides


class ProvidesKeyToAtomMappingImpl(AbstractProvidesKeyToAtomMapping):

    __slots__ = ()

    def __init__(self):
        pass

    @inject_items({
        "graph_mapper": "MemoryGraphMapper"
    })
    @overrides(
        AbstractProvidesKeyToAtomMapping.routing_key_partition_atom_mapping,
        additional_arguments={"graph_mapper"})
    def routing_key_partition_atom_mapping(
            self, routing_info, partition, graph_mapper):
        mapping = list()
        vertex_slice = graph_mapper.get_slice(partition.pre_vertex)
        keys = routing_info.get_keys(vertex_slice.n_atoms)
        atom = vertex_slice.lo_atom
        for key in keys:
            mapping.append((atom, key))
            atom += 1
        return mapping



import os
import hashlib
from spinn_front_end_common.utilities.constants import SDP_PORTS


def get_simulation_header_array(
        binary_file_name, machine_time_step, time_scale_factor):
    """ Get data to be written to the simulation header

    :param binary_file_name: The name of the binary of the application
    :param machine_time_step: The time step of the simulation
    :param time_scale_factor: The time scaling of the simulation
    :return: An array of values to be written as the simulation header
    """

    # Hash application title
    application_name = os.path.splitext(binary_file_name)[0]

    # Get first 32-bits of the md5 hash of the application name
    application_name_hash = hashlib.md5(application_name).hexdigest()[:8]

    # Write this to the system region (to be picked up by the simulation):
    data = list()
    data.append(int(application_name_hash, 16))
    data.append(machine_time_step * time_scale_factor)

    # add SDP port number for receiving synchronisations and new run times
    data.append(SDP_PORTS.RUNNING_COMMAND_SDP_PORT.value)

    return data



from data_specification import constants
from spinn_utilities.progress_bar import ProgressBar
from spinn_front_end_common.utilities.exceptions import ConfigurationException

import logging
import os
import struct

logger = logging.getLogger(__name__)

MEM_MAP_SUBDIR_NAME = "memory_map_reports"


class MemoryMapOnChipReport(object):
    """ Report on memory usage
    """

    def __call__(
            self, report_default_directory, dsg_targets, transceiver,
            loaded_app_data_token):
        """ creates a report that states where in sdram each region is

        :param report_default_directory: the folder where reports are written
        :param dsg_targets: the map between placement and file writer
        :param transceiver: the spinnMan instance
        :param loaded_app_data_token: flag that app data has been loaded
        :rtype: None
        """

        if not loaded_app_data_token:
            raise ConfigurationException(
                "Needs to have loaded app data for this to work.")

        directory_name = os.path.join(
            report_default_directory, MEM_MAP_SUBDIR_NAME)
        if not os.path.exists(directory_name):
            os.makedirs(directory_name)

        progress = ProgressBar(dsg_targets, "Writing memory map reports")
        for (x, y, p) in progress.over(dsg_targets):
            file_name = os.path.join(
                directory_name,
                "memory_map_from_processor"
                "_{0:d}_{1:d}_{2:d}.txt".format(x, y, p))
            with open(file_name, "w") as output:
                self._write_processor_memory_map(output, transceiver, x, y, p)

    def _write_processor_memory_map(self, output, txrx, x, y, p):
        output.write("On chip data specification executor\n\n")

        report_data_address = self._get_report_data_address(txrx, x, y, p)
        report_bytes = \
            _MemoryChannelState.STRUCT_SIZE * constants.MAX_MEM_REGIONS

        mem_map_report_data = buffer(txrx.read_memory(
            x, y, report_data_address, report_bytes))

        offset = 0
        for i in xrange(constants.MAX_MEM_REGIONS):
            region = _MemoryChannelState.from_bytestring(
                mem_map_report_data, offset)
            offset += _MemoryChannelState.STRUCT_SIZE

            if region.start_address == 0:
                output.write("Region {0:d}: Unused\n\n".format(i))
                continue

            output.write(
                "Region {0:d}:\n\t"
                "start address: 0x{1:x}\n\t"
                "size: {2:d}\n\t"
                "unfilled: {3:s}\n\t"
                "write pointer: 0x{4:x}\n\t"
                "size currently written(based on the "
                "write pointer): {5:d}\n\n".format(
                    i, region.start_address, region.size,
                    region.unfilled_tf, region.write_pointer,
                    0 if region.unfilled else region.written))

    @staticmethod
    def _get_report_data_address(txrx, x, y, p):
        data_address_pointer = txrx.get_user_1_register_address_from_core(
            x, y, p)
        data_address_encoded = txrx.read_memory(
            x, y, data_address_pointer, 4)
        return struct.unpack_from("<I", buffer(data_address_encoded))[0]


class _MemoryChannelState(object):
    # 4 fields each of 4 bytes
    STRUCT_SIZE = 16

    def __init__(self, start_address, size, unfilled, write_pointer):
        self._start_address = start_address
        self._size = size
        self._unfilled = unfilled
        self._write_pointer = write_pointer

    @property
    def start_address(self):
        return self._start_address

    @property
    def size(self):
        return self._size

    @property
    def unfilled(self):
        return self._unfilled

    @property
    def unfilled_tf(self):
        return "True" if self._unfilled else "False"

    @property
    def write_pointer(self):
        return self._write_pointer

    @property
    def written(self):
        return self._write_pointer - self._start_address

    @staticmethod
    def from_bytestring(data, offset=0):
        start, size, unfilled, write = \
            struct.unpack_from("<IIII", data, offset)
        return _MemoryChannelState(start, size, unfilled, write)

    def bytestring(self):
        return struct.pack("<IIII", self._start_address, self._size,
                           self._unfilled, self._write_pointer)



import os

from spinn_front_end_common.interface import interface_functions
from spinn_front_end_common.utilities import report_functions
from spinn_front_end_common import mapping_algorithms


def get_front_end_common_pacman_xml_paths():
    """ Get the XML path for the front end common interface functions
    """
    return [
        os.path.join(
            os.path.dirname(interface_functions.__file__),
            "front_end_common_interface_functions.xml"),
        os.path.join(
            os.path.dirname(report_functions.__file__),
            "front_end_common_reports.xml"),
        os.path.join(
            os.path.dirname(mapping_algorithms.__file__),
            "front_end_common_mapping_algorithms.xml"
        )
    ]



# spinn front end common imports
from pacman.model.resources.core_resource import CoreResource
from pacman.model.resources.pre_allocated_resource_container import \
    PreAllocatedResourceContainer
from pacman.model.resources.specific_chip_sdram_resource import \
    SpecificChipSDRAMResource

# spinn front end common imports
from spinn_front_end_common.utility_models.\
    chip_power_monitor_machine_vertex import \
    ChipPowerMonitorMachineVertex

# utils
from spinn_utilities.progress_bar import ProgressBar


class PreAllocateResourcesForChipPowerMonitor(object):
    """ Adds chip power monitor resources as required for a machine
    """

    def __call__(
            self, machine, n_machine_time_steps, n_samples_per_recording,
            sampling_frequency, time_scale_factor, machine_time_step,
            pre_allocated_resources=None):
        """
        :param pre_allocated_resources: other pre-allocated resources
        :param machine: the spinnaker machine as discovered
        :param n_machine_time_steps: the number of machine\
         time steps used by the simulation during this phase
        :param n_samples_per_recording: how many samples between record entries
        :param sampling_frequency: the frequency of sampling
        :param time_scale_factor: the time scale factor
        :param machine_time_step: the machine time step
        :return: pre allocated resources
        """

        progress_bar = ProgressBar(
            len(list(machine.chips)),
            "Preallocating resources for chip power monitor")

        # store how much SDRAM the LPG uses per core
        resources = ChipPowerMonitorMachineVertex.get_resources(
            n_machine_time_steps=n_machine_time_steps,
            n_samples_per_recording=n_samples_per_recording,
            sampling_frequency=sampling_frequency,
            time_scale_factor=time_scale_factor,
            time_step=machine_time_step)

        # for every Ethernet connected chip, get the resources needed by the
        # live packet gatherers
        sdrams = list()
        cores = list()
        for chip in progress_bar.over(machine.chips):
            sdrams.append(
                SpecificChipSDRAMResource(chip, resources.sdram.get_value()))
            cores.append(CoreResource(chip, 1))

        # create pre allocated resource container
        cpm_pre_allocated_resource_container = PreAllocatedResourceContainer(
            specific_sdram_usage=sdrams, core_resources=cores)

        # add other pre allocated resources
        if pre_allocated_resources is not None:
            cpm_pre_allocated_resource_container.extend(
                pre_allocated_resources)

        # return pre allocated resources
        return cpm_pre_allocated_resource_container



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


@add_metaclass(AbstractBase)
class AbstractGeneratesDataSpecification(object):

    __slots__ = ()

    @abstractmethod
    def generate_data_specification(self, spec, placement):
        """ Generate a data specification

        :param spec: The data specification to write to
        :param placement: the placement object this spec is associated with
        :type spec:\
            :py:class:`data_specification.DataSpecificationGenerator`
        :rtype: None
        """



from six import add_metaclass

from spinn_utilities.abstract_base import \
    AbstractBase, abstractmethod, abstractproperty


@add_metaclass(AbstractBase)
class AbstractChangableAfterRun(object):
    """ An item that can be changed after a call to run, the changes to which\
        might or might not require mapping
    """

    __slots__ = ()

    @abstractproperty
    def requires_mapping(self):
        """ True if changes that have been made require that mapping be\
            performed.  Note that this should return True the first time it\
            is called, as the vertex must require mapping as it has been\
            created!
        """

    @abstractmethod
    def mark_no_changes(self):
        """ Marks the point after which changes are reported.  Immediately\
            after calling this method, requires_mapping should return False.
        """



# dsg imports
from data_specification import utility_calls

# front end common imports
from spinn_front_end_common.utilities.exceptions import ConfigurationException

# SpiNMachine imports
from spinn_machine import CoreSubsets

# general imports
import os
import logging
import struct
import datetime
import shutil
from ConfigParser import RawConfigParser

logger = logging.getLogger(__name__)
FINISHED_FILENAME = "finished"


def read_data(x, y, address, length, data_format, transceiver):
    """ Reads and converts a single data item from memory

    :param x: chip x
    :param y: chip y
    :param address: base address of the SDRAM chip to read
    :param length: length to read
    :param data_format: the format to read memory
    :param transceiver: the SpinnMan interface
    """

    # turn byte array into str for unpack to work
    data = buffer(transceiver.read_memory(x, y, address, length))
    result = struct.unpack_from(data_format, data)[0]
    return result


def locate_memory_region_for_placement(placement, region, transceiver):
    """ Get the address of a region for a placement

    :param region: the region to locate the base address of
    :type region: int
    :param placement: the placement object to get the region address of
    :type placement: pacman.model.placements.Placement
    :param transceiver: the python interface to the spinnaker machine
    :type transceiver: spiNNMan.transciever.Transciever
    """
    regions_base_address = transceiver.get_cpu_information_from_core(
        placement.x, placement.y, placement.p).user[0]

    # Get the position of the region in the pointer table
    region_offset_in_pointer_table = \
        utility_calls.get_region_base_address_offset(
            regions_base_address, region)
    region_address = buffer(transceiver.read_memory(
        placement.x, placement.y, region_offset_in_pointer_table, 4))
    region_address_decoded = struct.unpack_from("<I", region_address)[0]
    return region_address_decoded


def set_up_output_application_data_specifics(
        where_to_write_application_data_files,
        max_application_binaries_kept, n_calls_to_run,
        this_run_time_string):
    """

    :param where_to_write_application_data_files:\
        the location where all app data is by default written to
    :param max_application_binaries_kept:\
        The max number of report folders to keep active at any one time
    :param n_calls_to_run: the counter of how many times run has been called.
    :param this_run_time_string: the time stamp string for this run
    :return: the run folder for this simulation to hold app data
    """
    this_run_time_folder = None
    if where_to_write_application_data_files == "DEFAULT":
        directory = os.getcwd()
        application_generated_data_file_folder = \
            os.path.join(directory, 'application_generated_data_files')
        if not os.path.exists(application_generated_data_file_folder):
            os.makedirs(application_generated_data_file_folder)

        _remove_excess_folders(
            max_application_binaries_kept,
            application_generated_data_file_folder)

        # add time stamped folder for this run
        this_run_time_folder = \
            os.path.join(
                application_generated_data_file_folder, this_run_time_string)
        if not os.path.exists(this_run_time_folder):
            os.makedirs(this_run_time_folder)

        # store timestamp in latest/time_stamp
        time_of_run_file_name = os.path.join(this_run_time_folder,
                                             "time_stamp")
        writer = open(time_of_run_file_name, "w")
        writer.writelines("{}".format(this_run_time_string))
        writer.flush()
        writer.close()

    else:
        # add time stamped folder for this run
        application_generated_data_file_folder = \
            os.path.join(where_to_write_application_data_files,
                         'application_generated_data_files')
        if not os.path.exists(application_generated_data_file_folder):
            os.makedirs(application_generated_data_file_folder)
        this_run_time_folder = \
            os.path.join(application_generated_data_file_folder,
                         this_run_time_string)
        if not os.path.exists(this_run_time_folder):
            os.makedirs(this_run_time_folder)

        # remove folders that are old and above the limit
        _remove_excess_folders(
            max_application_binaries_kept,
            where_to_write_application_data_files)

        # store timestamp in latest/time_stamp
        time_of_run_file_name = os.path.join(this_run_time_folder,
                                             "time_stamp")
        writer = open(time_of_run_file_name, "w")
        writer.writelines("{}".format(this_run_time_string))

        if not os.path.exists(this_run_time_folder):
            os.makedirs(this_run_time_folder)

    # create sub folder within reports for sub runs (where changes need to be
    # recorded)
    this_run_time_sub_folder = os.path.join(
        this_run_time_folder, "run_{}".format(n_calls_to_run))

    if not os.path.exists(this_run_time_sub_folder):
        os.makedirs(this_run_time_sub_folder)

    return this_run_time_sub_folder, this_run_time_folder


def set_up_report_specifics(
        default_report_file_path, max_reports_kept, n_calls_to_run,
        this_run_time_string=None):
    """

    :param default_report_file_path: The location where all reports reside
    :param max_reports_kept:\
        The max number of report folders to keep active at any one time
    :param n_calls_to_run: the counter of how many times run has been called.
    :param this_run_time_string: holder for the timestamp for future runs
    :return: The folder for this run, the time_stamp
    """

    # determine common report folder
    config_param = default_report_file_path
    created_folder = False
    if config_param == "DEFAULT":
        directory = os.getcwd()

        # global reports folder
        report_default_directory = os.path.join(directory, 'reports')
        if not os.path.exists(report_default_directory):
            os.makedirs(report_default_directory)
            created_folder = True
    elif config_param == "REPORTS":
        report_default_directory = 'reports'
        if not os.path.exists(report_default_directory):
            os.makedirs(report_default_directory)
    else:
        report_default_directory = \
            os.path.join(config_param, 'reports')
        if not os.path.exists(report_default_directory):
            os.makedirs(report_default_directory)

    # clear and clean out folders considered not useful anymore
    if not created_folder and len(os.listdir(report_default_directory)) > 0:
        _remove_excess_folders(max_reports_kept, report_default_directory)

    # determine the time slot for later
    if this_run_time_string is None:
        this_run_time = datetime.datetime.now()
        this_run_time_string = (
            "{:04}-{:02}-{:02}-{:02}-{:02}-{:02}-{:02}".format(
                this_run_time.year, this_run_time.month, this_run_time.day,
                this_run_time.hour, this_run_time.minute,
                this_run_time.second, this_run_time.microsecond))

    # handle timing app folder and cleaning of report folder from last run
    app_folder_name = os.path.join(
        report_default_directory, this_run_time_string)

    if not os.path.exists(app_folder_name):
            os.makedirs(app_folder_name)

    # create sub folder within reports for sub runs (where changes need to be
    # recorded)
    app_sub_folder_name = os.path.join(
        app_folder_name, "run_{}".format(n_calls_to_run))

    if not os.path.exists(app_sub_folder_name):
        os.makedirs(app_sub_folder_name)

    # store timestamp in latest/time_stamp for provenance reasons
    time_of_run_file_name = os.path.join(app_folder_name, "time_stamp")
    writer = open(time_of_run_file_name, "w")
    writer.writelines("{}".format(this_run_time_string))
    writer.flush()
    writer.close()
    return app_sub_folder_name, app_folder_name, this_run_time_string


def write_finished_file(app_data_runtime_folder, report_default_directory):
    # write a finished file that allows file removal to only remove folders
    # that are finished
    app_file_name = os.path.join(app_data_runtime_folder, FINISHED_FILENAME)
    writer = open(app_file_name, "w")
    writer.writelines("finished")
    writer.flush()
    writer.close()

    app_file_name = os.path.join(report_default_directory, FINISHED_FILENAME)
    writer = open(app_file_name, "w")
    writer.writelines("finished")
    writer.flush()
    writer.close()


def _remove_excess_folders(max_to_keep, starting_directory):
    files_in_report_folder = os.listdir(starting_directory)

    # while there's more than the valid max, remove the oldest one
    if len(files_in_report_folder) > max_to_keep:

        # sort files into time frame
        files_in_report_folder.sort(
            cmp, key=lambda temp_file:
            os.path.getmtime(os.path.join(starting_directory,
                                          temp_file)))

        # remove only the number of files required, and only if they have
        # the finished flag file created
        num_files_to_remove = len(files_in_report_folder) - max_to_keep
        files_removed = 0
        files_not_closed = 0
        for current_oldest_file in files_in_report_folder:
            finished_flag = os.path.join(os.path.join(
                starting_directory, current_oldest_file), FINISHED_FILENAME)
            if os.path.exists(finished_flag):
                shutil.rmtree(os.path.join(starting_directory,
                                           current_oldest_file),
                              ignore_errors=True)
                files_removed += 1
            else:
                files_not_closed += 1
            if (files_removed + files_not_closed) >= num_files_to_remove:
                break
        if files_not_closed > max_to_keep / 4:
            logger.warning("{} has {} old reports that have not been closed".
                           format(starting_directory, files_not_closed))


def convert_string_into_chip_and_core_subset(cores):
    """ Translate a string list of cores into a core subset

    :param cores:\
        string representing down cores formatted as x,y,p[:x,y,p]*
    :type cores: str or None
    """
    ignored_cores = CoreSubsets()
    if cores is not None and cores != "None":
        for downed_core in cores.split(":"):
            x, y, processor_id = downed_core.split(",")
            ignored_cores.add_processor(int(x), int(y), int(processor_id))
    return ignored_cores


def sort_out_downed_chips_cores_links(
        downed_chips, downed_cores, downed_links):
    """ Translate the down cores and down chips string into a form that \
        spinnman can understand

    :param downed_cores:\
        string representing down cores formatted as x,y,p[:x,y,p]*
    :type downed_cores: str or None
    :param downed_chips:\
        string representing down chips formatted as x,y[:x,y]*
    :type downed_chips: str or None
    :param downed_links:\
        string representing down links formatted as x,y,link[:x,y,link]*
    :return:\
        a tuple of (\
            set of (x, y) of down chips, \
            set of (x, y, p) of down cores, \
            set of ((x, y), link id) of down links)
    :rtype: ({(int, int,), }, {(int, int, int), }, {((int, int), int), })
    """
    ignored_chips = set()
    if downed_chips is not None and downed_chips != "None":
        for downed_chip in downed_chips.split(":"):
            x, y = downed_chip.split(",")
            ignored_chips.add((int(x), int(y)))

    ignored_cores = set()
    if downed_cores is not None and downed_cores != "None":
        for downed_core in downed_cores.split(":"):
            x, y, processor_id = downed_core.split(",")
            ignored_cores.add((int(x), int(y), int(processor_id)))

    ignored_links = set()
    if downed_links is not None and downed_links != "None":
        for downed_link in downed_links.split(":"):
            x, y, link_id = downed_link.split(",")
            ignored_links.add((int(x), int(y), int(link_id)))
    return ignored_chips, ignored_cores, ignored_links


def translate_iobuf_extraction_elements(
        hard_coded_cores, hard_coded_model_binary, executable_targets,
        executable_finder):
    """

    :param hard_coded_cores: list of cores to read iobuf from
    :param hard_coded_model_binary: list of binary names to read iobuf from
    :param executable_targets: the targets of cores and executable binaries
    :param executable_finder: where to find binaries paths from binary names
    :return: core subsets for the cores to read iobuf from
    """
    # all the cores
    if hard_coded_cores == "ALL" and hard_coded_model_binary == "None":
        return executable_targets.all_core_subsets

    # some hard coded cores
    if hard_coded_cores != "None" and hard_coded_model_binary == "None":
        ignored_cores = convert_string_into_chip_and_core_subset(
            hard_coded_cores)
        return ignored_cores

    # some binaries
    if hard_coded_cores == "None" and hard_coded_model_binary != "None":
        return _handle_model_binaries(
            hard_coded_model_binary, executable_targets, executable_finder)

    # nothing
    if hard_coded_cores == "None" and hard_coded_model_binary == "None":
        return CoreSubsets()

    # bit of both
    if hard_coded_cores != "None" and hard_coded_model_binary != "None":
        model_core_subsets = _handle_model_binaries(
            hard_coded_model_binary, executable_targets, executable_finder)
        hard_coded_core_core_subsets = \
            convert_string_into_chip_and_core_subset(hard_coded_cores)
        for core_subset in hard_coded_core_core_subsets:
            model_core_subsets.add_core_subset(core_subset)
        return model_core_subsets

    # should never get here,
    raise ConfigurationException("Something odd has happened")


def _handle_model_binaries(
        hard_coded_model_binary, executable_targets, executable_finder):
    """
    :param hard_coded_model_binary: list of binary names to read iobuf from
    :param executable_targets: the targets of cores and executable binaries
    :param executable_finder: where to find binaries paths from binary names
    :return: core subsets from binaries that need iobuf to be read from them
    """
    model_binaries = hard_coded_model_binary.split(",")
    cores = CoreSubsets()
    for model_binary in model_binaries:
        model_binary_path = \
            executable_finder.get_executable_path(model_binary)
        core_subsets = \
            executable_targets.get_cores_for_binary(model_binary_path)
        for core_subset in core_subsets:
            cores.add_core_subset(core_subset)
    return cores


def read_config(config, section, item):
    """ Get the string value of a config item, returning None if the value\
        is "None"
    """
    value = config.get(section, item)
    if value == "None":
        return None
    return value


def read_config_int(config, section, item):
    """ Get the integer value of a config item, returning None if the value\
        is "None"
    """
    value = read_config(config, section, item)
    if value is None:
        return value
    return int(value)


def read_config_boolean(config, section, item):
    """ Get the boolean value of a config item, returning None if the value\
        is "None"
    """
    value = read_config(config, section, item)
    if value is None:
        return value
    if value.lower() in RawConfigParser._boolean_states:
        return RawConfigParser._boolean_states[value.lower()]
    raise ValueError("Unknown boolean value {} in configuration {}:{}".format(
        value, section, item))


def generate_unique_folder_name(folder, filename, extension):
    """ Generate a unique file name with a given extension in a given folder

    :param folder: where to put this unique file
    :param filename: the name of the first part of the file without extension
    :param extension: extension of the file
    :return: file path with a unique addition
    """
    new_file_path = os.path.join(folder, "{}{}".format(filename, extension))
    count = 2
    while os.path.exists(new_file_path):
        new_file_path = os.path.join(
            folder, "{}_{}{}".format(filename, count, extension))
        count += 1
    return new_file_path


def get_ethernet_chip(machine, board_address):
    """ locate the chip with the given board IP address

    :param machine: the spinnaker machine
    :param board_address: the board address to locate the chip of.
    :return: The chip that supports that board address
    :raises ConfigurationException:\
        when that board address has no chip associated with it
    """
    for chip in machine.ethernet_connected_chips:
        if chip.ip_address == board_address:
            return chip
    raise ConfigurationException(
        "cannot find the Ethernet connected chip with the board address {}"
        .format(board_address))


def convert_time_diff_to_total_milliseconds(sample):
    """ converts between a time diff and total milliseconds

    :return: total milliseconds
    """
    return (sample.total_seconds() * 1000.0) + (sample.microseconds / 1000.0)



# spinnman imports
from spinnman.exceptions \
    import SpinnmanIOException, SpinnmanInvalidPacketException, \
    SpinnmanTimeoutException
from spinnman.messages.eieio.command_messages import EIEIOCommandHeader
from spinnman.connections.udp_packet_connections import UDPConnection
from spinnman.constants import EIEIO_COMMAND_IDS as CMDS

# FrontEndCommon imports
from .database_reader import DatabaseReader

# general imports
from threading import Thread
import logging

logger = logging.getLogger(__name__)


class DatabaseConnection(UDPConnection, Thread):
    """ A connection from the toolchain which will be notified\
        when the database has been written, and can then respond when the\
        database has been read, and further wait for notification that the\
        simulation has started.
    """

    def __init__(self, start_resume_callback_function=None,
                 stop_pause_callback_function=None, local_host=None,
                 local_port=19999):
        """

        :param start_resume_callback_function: A function to be called when
                    the start message has been received.  This function should
                    not take any parameters or return anything.
        :type start_resume_callback_function: function() -> None
        :param local_host: Optional specification of the local hostname or\
                    ip address of the interface to listen on
        :type local_host: str
        :param local_port: Optional specification of the local port to listen\
                    on.  Must match the port that the toolchain will send the\
                    notification on (19999 by default)
        :type local_port: int
        """
        UDPConnection.__init__(
            self, local_host=local_host, local_port=local_port,
            remote_host=None, remote_port=None)
        Thread.__init__(
            self, name="SpyNNakerDatabaseConnection:{}:{}".format(
                self.local_ip_address, self.local_port))
        self._database_callback_functions = list()
        self._start_resume_callback_function = start_resume_callback_function
        self._pause_and_stop_callback_function = stop_pause_callback_function
        self._running = False
        self.daemon = True
        self.start()

    def add_database_callback(self, database_callback_function):
        """ Add a database callback to be called when the database is ready

        :param database_callback_function: A function to be called when the\
            database message has been received.  This function should take \
            a single parameter, which will be a DatabaseReader object. \
            Once the function returns, it will be assumed that the database \
            has been read, and the return response will be sent.
        :type database_callback_function: function(\
            :py:class:`spynnaker_external_devices.pyNN.connections.database_reader.DatabaseReader`)\
                    -> None
        """
        self._database_callback_functions.append(database_callback_function)

    def run(self):
        self._running = True
        logger.info(
            "{}:{} Waiting for message to indicate that the database is "
            "ready".format(self.local_ip_address, self.local_port))
        while self._running:
            try:
                data, address = self._retrieve_database_address()
                if data is not None:
                    self._process_message(address, data)
            except Exception as e:
                logger.error("Failure processing database callback",
                             exc_info=True)
                raise SpinnmanIOException(str(e))

    def _process_message(self, address, data):
        # Read the read packet confirmation
        logger.info("{}:{} Reading database".format(
            self.local_ip_address, self.local_port))
        database_path = str(data[2:])

        # Call the callback
        database_reader = DatabaseReader(database_path)
        for database_callback in self._database_callback_functions:
            database_callback(database_reader)
        database_reader.close()

        # Send the response
        logger.info("Notifying the toolchain that the database has been read")
        self._send_command(1, address)

        # Wait for the start of the simulation
        if self._start_resume_callback_function is not None:
            logger.info(
                "Waiting for message to indicate that the simulation has "
                "started or resumed")
            command_code = self._receive_command()
            if command_code != CMDS.START_RESUME_NOTIFICATION.value:
                raise SpinnmanInvalidPacketException(
                    "command_code",
                    "expected a start/resume command code now, and did not "
                    "receive it")
            # Call the callback
            self._start_resume_callback_function()

        if self._pause_and_stop_callback_function is not None:
            logger.info(
                "Waiting for message to indicate that the simulation has "
                "stopped or paused")
            command_code = self._receive_command()
            if command_code != CMDS.STOP_PAUSE_NOTIFICATION.value:
                raise SpinnmanInvalidPacketException(
                    "command_code",
                    "expected a pause/stop command code now, and did not "
                    "receive it")
            # Call the callback
            self._pause_and_stop_callback_function()

    def _send_command(self, command, address):
        self.send_to(EIEIOCommandHeader(command).bytestring, address)

    def _receive_command(self):
        return EIEIOCommandHeader.from_bytestring(self.receive(), 0).command

    def _retrieve_database_address(self):
        try:
            data, address = self.receive_with_address(timeout=3)
            return data, address
        except SpinnmanTimeoutException:
            return None, None
        except SpinnmanIOException:
            raise

    def close(self):
        self._running = False
        UDPConnection.close(self)



# utilities imports
from spinn_utilities.progress_bar import ProgressBar

# front end common imports
from spinn_front_end_common.utilities.database import DatabaseWriter


class DatabaseInterface(object):
    """ Writes a database of the graph(s) and other information
    """

    __slots__ = [
        # the database writer object
        "_writer",

        # True if the end user has asked for the database to be written
        "_user_create_database",

        # True if the network is computed to need the database to be written
        "_needs_database"
    ]

    def __init__(self):
        self._writer = None
        self._user_create_database = None
        self._needs_database = None

    def __call__(
            self, machine_graph, user_create_database, tags,
            runtime, machine, time_scale_factor, machine_time_step,
            placements, routing_infos, router_tables, database_directory,
            create_atom_to_event_id_mapping=False, application_graph=None,
            graph_mapper=None):

        self._writer = DatabaseWriter(database_directory)
        self._user_create_database = user_create_database
        # add database generation if requested
        self._needs_database = self._writer.auto_detect_database(
            machine_graph)

        if self.needs_database:
            with self._writer as w, ProgressBar(9, "Creating database") as p:
                w.add_system_params(
                    time_scale_factor, machine_time_step, runtime)
                p.update()
                w.add_machine_objects(machine)
                p.update()
                if application_graph is not None \
                        and application_graph.n_vertices:
                    w.add_application_vertices(application_graph)
                p.update()
                w.add_vertices(machine_graph, graph_mapper, application_graph)
                p.update()
                w.add_placements(placements)
                p.update()
                w.add_routing_infos(routing_infos, machine_graph)
                p.update()
                w.add_routing_tables(router_tables)
                p.update()
                w.add_tags(machine_graph, tags)
                p.update()
                if (graph_mapper is not None
                        and application_graph is not None
                        and create_atom_to_event_id_mapping):
                    w.create_atom_to_event_id_mapping(
                        graph_mapper=graph_mapper,
                        application_graph=application_graph,
                        machine_graph=machine_graph,
                        routing_infos=routing_infos)
                p.update()

        return self, self.database_file_path

    @property
    def needs_database(self):
        return ((self._user_create_database == "None" and self._needs_database)
                or self._user_create_database == "True")

    @property
    def database_file_path(self):
        if self.needs_database:
            return self._writer.database_path
        return None



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


@add_metaclass(AbstractBase)
class AbstractMachineAllocationController(object):
    """ An object that controls the allocation of a machine
    """

    __slots__ = ()

    @abstractmethod
    def extend_allocation(self, new_total_run_time):
        """ Extend the allocation of the machine from the original\
            run time

        :param new_total_run_time: The total run time that is now required\
                    starting from when the machine was first allocated
        """

    @abstractmethod
    def close(self):
        """ Indicate that the use of the machine is complete
        """



# spinn_machine imports
from spinn_machine import VirtualMachine


class VirtualMachineGenerator(object):

    __slots__ = []

    def __call__(
            self, width=None, height=None, virtual_has_wrap_arounds=None,
            version=None, n_cpus_per_chip=18, with_monitors=True,
            down_chips=None, down_cores=None, down_links=None):
        """
        :param width: The width of the machine in chips
        :param height: The height of the machine in chips
        :param virtual_has_wrap_arounds: True if the machine\
                should be created with wrap_arounds
        :param version: The version of board to create
        :param n_cpus_per_chip: The number of cores to put on each chip
        :param with_monitors: If true, CPU 0 will be marked as a monitor
        """

        machine = VirtualMachine(
            width=width, height=height,
            with_wrap_arounds=virtual_has_wrap_arounds,
            version=version, n_cpus_per_chip=n_cpus_per_chip,
            with_monitors=with_monitors, down_chips=down_chips,
            down_cores=down_cores, down_links=down_links)

        # Work out and add the spinnaker links and FPGA links
        machine.add_spinnaker_links(version)
        machine.add_fpga_links(version)

        return machine



import sqlite3


class DatabaseReader(object):
    """ A reader for the database
    """

    __slots__ = [
        # the database connection (is basically a lock on the database)
        "_connection",

        # the location for writing
        "_cursor"
    ]

    def __init__(self, database_path):
        """

        :param database_path: The path to the database
        :type database_path: str
        """
        self._connection = sqlite3.connect(database_path)
        self._connection.row_factory = sqlite3.Row
        self._cursor = self._connection.cursor()

    @property
    def cursor(self):
        """ The database cursor.  Allows custom SQL queries to be performed.

        :rtype: :py:class:`sqlite3.Cursor`
        """
        return self._cursor

    def get_key_to_atom_id_mapping(self, label):
        """ Get a mapping of event key to atom id for a given vertex

        :param label: The label of the vertex
        :type label: str
        :return: dictionary of atom ids indexed by event key
        :rtype: dict
        """
        event_id_to_atom_id_mapping = dict()
        for row in self._cursor.execute(
                "SELECT n.atom_id AS a_id, n.event_id AS event"
                " FROM event_to_atom_mapping AS n"
                " JOIN Application_vertices AS p ON n.vertex_id = p.vertex_id"
                " WHERE p.vertex_label = ?", (label, )):
            event_id_to_atom_id_mapping[row["event"]] = row["a_id"]
        return event_id_to_atom_id_mapping

    def get_atom_id_to_key_mapping(self, label):
        """ Get a mapping of atom id to event key for a given vertex

        :param label: The label of the vertex
        :type label: str
        :return: dictionary of event keys indexed by atom id
        """
        atom_to_event_id_mapping = dict()
        for row in self._cursor.execute(
                "SELECT n.atom_id AS a_id, n.event_id AS event"
                " FROM event_to_atom_mapping AS n"
                " JOIN Application_vertices AS p"
                "   ON n.vertex_id = p.vertex_id"
                " WHERE p.vertex_label = ?", (label, )):
            atom_to_event_id_mapping[row["a_id"]] = row["event"]
        return atom_to_event_id_mapping

    def get_live_output_details(self, label, receiver_label):
        """ Get the ip address, port and whether the SDP headers are to be\
            stripped from the output from a vertex

        :param label: The label of the vertex
        :type label: str
        :return: tuple of (ip address, port, strip SDP)
        :rtype: (str, int, bool)
        """
        self._cursor.execute(
            "SELECT * FROM IP_tags AS tag"
            " JOIN graph_mapper_vertex AS mapper"
            "   ON tag.vertex_id = mapper.machine_vertex_id"
            " JOIN Application_vertices AS post_vertices"
            "   ON mapper.application_vertex_id = post_vertices.vertex_id"
            " JOIN Application_edges AS edges"
            "   ON mapper.application_vertex_id = edges.post_vertex"
            " JOIN Application_vertices AS pre_vertices"
            "   ON edges.pre_vertex = pre_vertices.vertex_id"
            " WHERE pre_vertices.vertex_label = ?"
            "   AND post_vertices.vertex_label = ?"
            " LIMIT 1", (label, receiver_label))
        row = self._cursor.fetchone()
        return (
            row["ip_address"], row["port"], row["strip_sdp"],
            row["board_address"])

    def get_live_input_details(self, label):
        """ Get the ip address and port where live input should be sent\
            for a given vertex

        :param label: The label of the vertex
        :type label: str
        :return: tuple of (ip address, port)
        :rtype: (str, int)
        """
        self._cursor.execute(
            "SELECT tag.board_address, tag.port AS port"
            " FROM Reverse_IP_tags AS tag"
            " JOIN graph_mapper_vertex AS mapper"
            "   ON tag.vertex_id = mapper.machine_vertex_id"
            " JOIN Application_vertices AS application"
            "   ON mapper.application_vertex_id = application.vertex_id"
            " WHERE application.vertex_label = ?"
            " LIMIT 1", (label, ))
        row = self._cursor.fetchone()
        return row["board_address"], row["port"]

    def get_machine_live_output_details(self, label, receiver_label):
        """ Get the ip address, port and whether the SDP headers are to be\
            stripped from the output from a machine vertex

        :param label: The label of the vertex
        :type label: str
        :return: tuple of (ip address, port, strip SDP)
        :rtype: (str, int, bool)
        """
        self._cursor.execute(
            "SELECT * FROM IP_tags AS tag"
            " JOIN Machine_vertices AS post_vertices"
            "   ON tag.vertex_id = post_vertices.vertex_id"
            " JOIN Machine_edges AS edges"
            "   ON post_vertices.vertex_id = edges.post_vertex"
            " JOIN Machine_vertices AS pre_vertices"
            "   ON edges.pre_vertex = pre_vertices.vertex_id"
            " WHERE pre_vertices.label = ?"
            "   AND post_vertices.label = ?"
            " LIMIT 1", (label, receiver_label))
        row = self._cursor.fetchone()
        return (
            row["ip_address"], row["port"], row["strip_sdp"],
            row["board_address"])

    def get_machine_live_input_details(self, label):
        """ Get the ip address and port where live input should be sent\
            for a given machine vertex

        :param label: The label of the vertex
        :type label: str
        :return: tuple of (ip address, port)
        :rtype: (str, int)
        """
        self._cursor.execute(
            "SELECT tag.board_address, tag.port AS port"
            " FROM Reverse_IP_tags AS tag"
            " JOIN Machine_vertices AS post_vertices"
            "   ON tag.vertex_id = post_vertices.vertex_id"
            " WHERE post_vertices.label = ?"
            " LIMIT 1", (label, ))
        row = self._cursor.fetchone()
        return row["board_address"], row["port"]

    def get_machine_live_output_key(self, label, receiver_label):
        self._cursor.execute(
            "SELECT * FROM Routing_info AS r_info"
            " JOIN Machine_edges AS edges"
            "   ON edges.edge_id = r_info.edge_id"
            " JOIN Machine_vertices AS post_vertices"
            "   ON post_vertices.vertex_id = edges.post_vertex"
            " JOIN Machine_vertices AS pre_vertices"
            "   ON pre_vertices.vertex_id = edges.pre_vertex"
            " WHERE pre_vertices.label = ?"
            "   AND post_vertices.label = ?"
            " LIMIT 1", (label, receiver_label))
        row = self._cursor.fetchone()
        return (row["key"], row["mask"])

    def get_machine_live_input_key(self, label):
        self._cursor.execute(
            "SELECT * FROM Routing_info AS r_info"
            " JOIN Machine_edges AS edges"
            "   ON edges.edge_id = r_info.edge_id"
            " JOIN Machine_vertices AS pre_vertices"
            "   ON pre_vertices.vertex_id = edges.pre_vertex"
            " WHERE pre_vertices.label = ?"
            " LIMIT 1", (label, ))
        row = self._cursor.fetchone()
        return (row["key"], row["mask"])

    def get_n_atoms(self, label):
        """ Get the number of atoms in a given vertex

        :param label: The label of the vertex
        :type label: str
        :return: The number of atoms
        :rtype: int
        """
        self._cursor.execute(
            "SELECT no_atoms FROM Application_vertices "
            "WHERE vertex_label = ?"
            " LIMIT 1", (label, ))
        return self._cursor.fetchone()["no_atoms"]

    def get_configuration_parameter_value(self, parameter_name):
        """ Get the value of a configuration parameter

        :param parameter_name: The name of the parameter
        :type parameter_name: str
        :return: The value of the parameter
        :rtype: float
        """
        self._cursor.execute(
            "SELECT value FROM configuration_parameters"
            " WHERE parameter_id = ?"
            " LIMIT 1", (parameter_name, ))
        return float(self._cursor.fetchone()["value"])

    def close(self):
        self._connection.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # @UnusedVariable
        self._connection.close()
        return False



import requests


class HBPMaxMachineGenerator(object):

    """ Generates the width and height of the maximum machine a given\
        HBP server can generate
    """

    __slots__ = []

    def __call__(self, hbp_server_url, total_run_time):
        """

        :param hbp_server_url: The URL of the HBP server from which to get\
                    the machine
        :param total_run_time: The total run time to request
        """

        max_machine = self._max_machine_request(hbp_server_url, total_run_time)

        # Return the width and height and assume that it has wrap arounds
        return max_machine["width"], max_machine["height"], True

    def _max_machine_request(self, url, total_run_time):
        if url.endswith("/"):
            url = url[:-1]
        return requests.get(
            "{}/max".format(url),
            params={'runTime': total_run_time}).json()



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase
from spinn_utilities.abstract_base import abstractproperty


@add_metaclass(AbstractBase)
class AbstractSendMeMulticastCommandsVertex(object):
    """ A vertex which wants to commands to be sent to it as multicast packets
        at fixed points in the simulation
    """

    __slots__ = ()

    @abstractproperty
    def start_resume_commands(self):
        """ The commands needed when starting or resuming simulation
        """

    @abstractproperty
    def pause_stop_commands(self):
        """ The commands needed when pausing or stopping simulation
        """

    @abstractproperty
    def timed_commands(self):
        """ The commands to be sent at given times in the simulation
        """



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


@add_metaclass(AbstractBase)
class AbstractProvidesOutgoingPartitionConstraints(object):
    """ A vertex that can provide constraints for its outgoing edge partitions
    """

    __slots__ = ()

    @abstractmethod
    def get_outgoing_partition_constraints(self, partition):
        """ Get constraints to be added to the given edge that comes out of\
            this vertex

        :param partition: An edge that comes out of this vertex
        :return: A list of constraints
        :rtype: list of\
                    :py:class:`pacman.model.constraints.abstract_constraint.AbstractConstraint`
        """
        pass



# pacman imports
from pacman.model.routing_info \
    import DictBasedMachinePartitionNKeysMap

# utilities imports
from spinn_utilities.progress_bar import ProgressBar

# front end common imports
from spinn_front_end_common.abstract_models import \
    AbstractProvidesIncomingPartitionConstraints, \
    AbstractProvidesNKeysForPartition, \
    AbstractProvidesOutgoingPartitionConstraints
from spinn_front_end_common.utilities.exceptions import ConfigurationException


class EdgeToNKeysMapper(object):
    """ Works out the number of keys needed for each edge
    """

    __slots__ = []

    def __call__(self, machine_graph=None, application_graph=None,
                 graph_mapper=None):
        # Generate an n_keys map for the graph and add constraints
        n_keys_map = DictBasedMachinePartitionNKeysMap()

        if machine_graph is None:
            raise ConfigurationException(
                "A machine graph is required for this mapper. "
                "Please choose and try again")
        if (application_graph is None) != (graph_mapper is None):
            raise ConfigurationException(
                "Can only do one graph. semantically doing 2 graphs makes no "
                "sense. Please choose and try again")

        if application_graph is not None:
            # generate progress bar
            progress = ProgressBar(
                machine_graph.n_vertices,
                "Getting number of keys required by each edge using "
                "application graph")

            # iterate over each partition in the graph
            for vertex in progress.over(machine_graph.vertices):
                partitions = machine_graph.\
                    get_outgoing_edge_partitions_starting_at_vertex(
                        vertex)
                for partition in partitions:
                    added_constraints = False
                    constraints = self._process_application_partition(
                        partition, n_keys_map, graph_mapper)
                    if not added_constraints:
                        partition.add_constraints(constraints)
                    else:
                        self._check_constraints_equal(
                            constraints, partition.constraints)

        else:
            # generate progress bar
            progress = ProgressBar(
                machine_graph.n_vertices,
                "Getting number of keys required by each edge using "
                "machine graph")

            for vertex in progress.over(machine_graph.vertices):
                partitions = machine_graph.\
                    get_outgoing_edge_partitions_starting_at_vertex(
                        vertex)
                for partition in partitions:
                    added_constraints = False
                    constraints = self._process_machine_partition(
                        partition, n_keys_map)
                    if not added_constraints:
                        partition.add_constraints(constraints)
                    else:
                        self._check_constraints_equal(
                            constraints, partition.constraints)

        return n_keys_map

    @staticmethod
    def _check_constraints_equal(constraints, stored_constraints):
        """

        :param constraints:
        :param stored_constraints:
        :rtype: None
        """
        for constraint in constraints:
            if constraint not in stored_constraints:
                raise ConfigurationException(
                    "Two edges within the same partition have different "
                    "constraints")

    @staticmethod
    def _process_application_partition(partition, n_keys_map, graph_mapper):
        vertex_slice = graph_mapper.get_slice(
            partition.pre_vertex)
        vertex = graph_mapper.get_application_vertex(
            partition.pre_vertex)

        if not isinstance(vertex, AbstractProvidesNKeysForPartition):
            n_keys_map.set_n_keys_for_partition(
                partition, vertex_slice.n_atoms)
        else:
            n_keys_map.set_n_keys_for_partition(
                partition,
                vertex.get_n_keys_for_partition(
                    partition, graph_mapper))

        constraints = list()
        if isinstance(vertex,
                      AbstractProvidesOutgoingPartitionConstraints):
            constraints.extend(
                vertex.get_outgoing_partition_constraints(partition))
        for edge in partition.edges:
            app_edge = graph_mapper.get_application_edge(edge)
            if isinstance(app_edge.post_vertex,
                          AbstractProvidesIncomingPartitionConstraints):
                constraints.extend(
                    app_edge.post_vertex.get_incoming_partition_constraints(
                        partition))
        constraints.extend(partition.constraints)
        return constraints

    @staticmethod
    def _process_machine_partition(partition, n_keys_map):

        if not isinstance(partition.pre_vertex,
                          AbstractProvidesNKeysForPartition):
            n_keys_map.set_n_keys_for_partition(partition, 1)
        else:
            n_keys_map.set_n_keys_for_partition(
                partition,
                partition.pre_vertex.get_n_keys_for_partition(partition, None))

        constraints = list()
        if isinstance(partition.pre_vertex,
                      AbstractProvidesOutgoingPartitionConstraints):
            constraints.extend(
                partition.pre_vertex.get_outgoing_partition_constraints(
                    partition))

        for edge in partition.edges:
            if isinstance(edge.post_vertex,
                          AbstractProvidesIncomingPartitionConstraints):
                constraints.extend(
                    edge.post_vertex.get_incoming_partition_constraints(
                        partition))
        constraints.extend(partition.constraints)

        return constraints



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractmethod


@add_metaclass(AbstractBase)
class AbstractProvidesLocalProvenanceData(object):
    """ Indicates an object that provides locally obtained provenance data
    """

    __slots__ = ()

    @abstractmethod
    def get_local_provenance_data(self):
        """ Get an iterable of provenance data items

        :return: iterable of ProvenanceDataItem
        """



from spinn_front_end_common._version import __version__  # NOQA
from spinn_front_end_common._version import __version_name__  # NOQA
from spinn_front_end_common._version import __version_month__  # NOQA
from spinn_front_end_common._version import __version_year__  # NOQA
__author__ = 'stokesa6'



# general imports
import logging
from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractmethod

logger = logging.getLogger(__name__)


@add_metaclass(AbstractBase)
class AbstractSendsBuffersFromHost(object):
    """ Interface to an object that sends buffers of keys to be\
        transmitted at given timestamps in the simulation
    """

    __slots__ = ()

    @abstractmethod
    def buffering_input(self):
        """ Return True if the input of this vertex is to be buffered
        """

    @abstractmethod
    def get_regions(self):
        """ Get the set of regions for which there are keys to be sent

        :return: Iterable of region ids
        :rtype: iterable of int
        """

    @abstractmethod
    def get_region_buffer_size(self, region):
        """ Get the size of the buffer to be used in SDRAM on the machine\
            for the region in bytes

        :param region: The region to get the buffer size of
        :type region: int
        :return: The size of the buffer space in bytes
        :rtype: int
        """

    @abstractmethod
    def get_max_buffer_size_possible(self, region):
        """ Gets the max possible size of a buffered region

        :param region: the region to find the max possible size of
        :type region: int
        :return: the max possible size of the buffered region
        :rtype: int
        """

    @abstractmethod
    def is_next_timestamp(self, region):
        """ Determine if there is another timestamp with data to be sent

        :param region: The region to determine if there is more data for
        :type region: int
        :return: True if there is more data, False otherwise
        :rtype: int
        """

    @abstractmethod
    def get_next_timestamp(self, region):
        """ Get the next timestamp at which there are still keys to be sent\
            for the given region

        :param region: The region to get the timestamp for
        :type region: int
        :return: The timestamp of the next available keys
        :rtype: int
        """

    @abstractmethod
    def is_next_key(self, region, timestamp):
        """ Determine if there are still keys to be sent at the given\
            timestamp for the given region

        :param region: The region to determine if there are keys for
        :type region: int
        :param timestamp: The timestamp to determine if there are more keys for
        :type timestamp: int
        :return: True if there are more keys to send for the parameters, False\
                    otherwise
        :rtype: bool
        """

    @abstractmethod
    def get_next_key(self, region):
        """ Get the next key in the given region

        :param region: The region to get the next key from
        :type region: int
        :return: The next key, or None if there are no more keys
        :rtype: int
        """

    @abstractmethod
    def is_empty(self, region):
        """ Return true if there are no spikes to be buffered for the\
            specified region

        :param region: The region to get the next key from
        :type region: int
        :return: True if there are no keys to send for the region, False\
                    otherwise
        :rtype: bool
        """

    @abstractmethod
    def rewind(self, region):
        """ Rewinds the internal buffer in preparation of re-sending
            the spikes

        :param region: The region to rewind
        :type region: int
        """



from spinn_front_end_common.interface.buffer_management.storage_objects\
    import ChannelBufferState
from spinn_front_end_common.utilities.constants \
    import SARK_PER_MALLOC_SDRAM_USAGE, SDP_PORTS

from pacman.model.resources \
    import ResourceContainer, IPtagResource, SDRAMResource

import struct
import sys
import math

# The offset of the last sequence number field in bytes
_LAST_SEQUENCE_NUMBER_OFFSET = 4 * 6

# The offset of the memory addresses in bytes
_FIRST_REGION_ADDRESS_OFFSET = 4 * 7

# the number of data elements inside the recording region before
# recording regions sizes are stored.
_RECORDING_ELEMENTS_BEFORE_REGION_SIZES = 7

# The Buffer traffic type
TRAFFIC_IDENTIFIER = "BufferTraffic"


def get_recording_header_size(n_recorded_regions):
    """ Get the size of the data to be written for the recording header

    :param n_recorded_regions: The number of regions to be recorded
    """

    # See recording.h/recording_initialise for data included in the header
    return (_RECORDING_ELEMENTS_BEFORE_REGION_SIZES +
            (2 * n_recorded_regions)) * 4


def get_recording_data_size(recorded_region_sizes):
    """ Get the size of the recorded data to be reserved

    :param recorded_region_sizes:\
        A list of sizes of each region to be recorded.\
        A size of 0 is acceptable.
    :rtype: int
    """
    return (
        # The total recording data size
        sum(recorded_region_sizes) +

        # The storage of the recording state
        (len(recorded_region_sizes) *
         ChannelBufferState.size_of_channel_state()) +

        # The SARK allocation of SDRAM overhead
        (len(recorded_region_sizes) * SARK_PER_MALLOC_SDRAM_USAGE))


def get_minimum_buffer_sdram(
        buffered_sdram_per_timestep, n_machine_time_steps=None,
        minimum_sdram_for_buffering=(1024 * 1024)):
    """ Get the minimum buffer SDRAM

    :param buffered_sdram_per_timestep:\
        The maximum number of bytes to use per timestep of recording,\
        per recorded region.  Disabled regions can specify 0.
    :type buffered_sdram_per_timestep: list of int
    :param n_machine_time_steps:\
        The number of machine time steps for the simulation.  Can be None if\
        use_auto_pause_and_resume is True
    :type n_machine_time_steps: int
    :param minimum_sdram_for_buffering:\
        The minimum SDRAM to reserve per recorded region for buffering
    :type minimum_sdram_for_buffering: int
    :rtype: list of int
    """

    # The minimum SDRAM for each region is:
    # - If the buffered_sdram_per_timestep for the region is > 0 and
    #   n_machine_time_steps is defined then the minimum of the actual region
    #   size and the minimum_sdram_for_buffering
    # - If the sdram is 0 then 0
    # - If n_machine_time_steps is None then minimum_sdram_for_buffering
    return [
        min(sdram * n_machine_time_steps, minimum_sdram_for_buffering)
        if sdram > 0 and n_machine_time_steps is not None
        else 0 if sdram == 0 else minimum_sdram_for_buffering
        for sdram in buffered_sdram_per_timestep
    ]


def get_recording_region_sizes(
        buffered_sdram_per_timestep, n_machine_time_steps=None,
        minimum_sdram_for_buffering=(1024 * 1024),
        maximum_sdram_for_buffering=None, use_auto_pause_and_resume=True):
    """ Get the size of each recording region to be passed in to\
        get_recording_resources, based on the details of the simulation

    :param buffered_sdram_per_timestep:\
        The maximum number of bytes to use per timestep of recording,\
        per recorded region.  Disabled regions can specify 0.
    :type buffered_sdram_per_timestep: list of int
    :param n_machine_time_steps:\
        The number of machine time steps for the simulation.  Can be None if\
        use_auto_pause_and_resume is True
    :type n_machine_time_steps: int
    :param minimum_sdram_for_buffering:\
        The minimum SDRAM to reserve per recorded region for buffering
    :type minimum_sdram_for_buffering: int
    :param maximum_sdram_for_buffering:\
        The maximum size of each buffer, or None if no maximum
    :type maximum_sdram_for_buffering: None or list of int
    :param use_auto_pause_and_resume:\
        True if automatic pause and resume is to be used for buffering
    :type use_auto_pause_and_resume: bool
    :rtype: list of int
    """
    if use_auto_pause_and_resume:

        # If auto pause and resume is enabled, find the minimum sizes
        return get_minimum_buffer_sdram(
            buffered_sdram_per_timestep, n_machine_time_steps,
            minimum_sdram_for_buffering)
    else:

        # If auto pause and resume is disabled, use the actual region size
        return get_recorded_region_sizes(
            n_machine_time_steps, buffered_sdram_per_timestep,
            maximum_sdram_for_buffering)


def get_recording_resources(
        region_sizes, buffering_ip_address=None,
        buffering_port=None, notification_tag=None):
    """ Get the resources for recording

    :param region_sizes:\
        A list of the sizes of each region.  A size of 0 is acceptable to\
        indicate an empty region
    :type region_sizes: list of int
    :param buffering_ip_address:\
        The ip address to receive buffering messages on, or None if buffering\
        is not in use
    :type buffering_ip_address: str
    :param buffering_port:\
        The port to receive buffering messages on, or None if a port is to be\
        assigned
    :type buffering_port: int
    :param notification_tag:\
        The tag to send buffering messages with, or None to use a default tag
    :type notification_tag: int
    :rtype:\
        :py:class:`pacman.model.resources.ResourceContainer`
    """

    ip_tags = list()
    if buffering_ip_address is not None:
        ip_tags.append(IPtagResource(
            buffering_ip_address, buffering_port, True, notification_tag,
            TRAFFIC_IDENTIFIER
        ))

    # return the resources including the SDRAM requirements
    return ResourceContainer(
        iptags=ip_tags,
        sdram=SDRAMResource(
            get_recording_header_size(len(region_sizes)) +
            get_recording_data_size(region_sizes)))


def get_recorded_region_sizes(
        n_machine_time_steps, buffered_sdram_per_timestep,
        maximum_sdram_for_buffering=None):
    """ Get the size of each recording region to be passed in to\
        get_recording_header_array

    :param n_machine_time_steps:\
        The duration of the simulation segment in time steps
    :type n_machine_time_steps: int
    :param buffered_sdram_per_timestep:\
        The maximum SDRAM used per timestep in bytes per region
    :type buffered_sdram_per_timestep: list of int
    :param maximum_sdram_for_buffering:\
        The maximum size of each buffer, or None if no maximum
    :type maximum_sdram_for_buffering: None or list of int
    :rtype: list of int
    """

    # The size of each buffer is the actual size needed for the number of
    # timesteps, or the maximum for buffering if a maximum is specified
    if n_machine_time_steps is None:
        data = [0 for _ in buffered_sdram_per_timestep]
        return data
    return [
        n_machine_time_steps * sdram
        if (maximum_sdram_for_buffering is None or
            maximum_sdram_for_buffering[i] == 0 or
            (n_machine_time_steps * sdram) < maximum_sdram_for_buffering[i])
        else maximum_sdram_for_buffering[i]
        for i, sdram in enumerate(buffered_sdram_per_timestep)
    ]


def get_recording_header_array(
        recorded_region_sizes,
        time_between_triggers=0, buffer_size_before_request=None, ip_tags=None,
        buffering_tag=None):
    """ Get data to be written for the recording header

    :param recorded_region_sizes:\
        A list of sizes of each region to be recorded.\
        A size of 0 is acceptable.
    :param time_between_triggers:\
        The minimum time between requesting reads of any region
    :param buffer_size_before_request:\
        The amount of buffer to fill before a read request is sent
    :param ip_tags: A list of ip tags to extract the buffer tag from
    :param buffering_tag: The tag to use for buffering requests
    :return: An array of values to be written as the header
    :rtype: list of int
    """

    # Find the tag if required
    buffering_output_tag = 0
    buffering_output_dest_x = 0
    buffering_output_dest_y = 0
    if buffering_tag is not None:
        buffering_output_tag = buffering_tag
    elif ip_tags is not None and len(ip_tags) > 0:
        buffering_output_tag = None
        for tag in ip_tags:
            if tag.traffic_identifier == TRAFFIC_IDENTIFIER:
                buffering_output_tag = tag.tag
                buffering_output_dest_x = tag.destination_x
                buffering_output_dest_y = tag.destination_y
                break
        if buffering_output_tag is None:
            raise Exception("Buffering tag not found")

    # See recording.h/recording_initialise for data included in the header
    data = list()

    # The parameters
    data.append(len(recorded_region_sizes))
    data.append(buffering_output_tag)
    data.append(struct.unpack("<I", struct.pack(
        "<HH", buffering_output_dest_y, buffering_output_dest_x))[0])
    data.append(SDP_PORTS.OUTPUT_BUFFERING_SDP_PORT.value)
    if buffer_size_before_request is not None:
        data.append(buffer_size_before_request)
    else:

        # If no buffer size before request, assume not buffering, so ensure
        # that the buffering will not be activated
        data.append(max(recorded_region_sizes) + 256)
    data.append(time_between_triggers)

    # The last sequence number (to be filled in by C code)
    data.append(0)

    # The pointers for each region (to be filled in by C code)
    data.extend([0 for _ in recorded_region_sizes])

    # The size of the regions
    data.extend(recorded_region_sizes)

    return data


def get_last_sequence_number(placement, transceiver, recording_data_address):
    """ Read the last sequence number from the data

    :param placement: The placement from which to read the sequence number
    :param transceiver: The transceiver to use to read the sequence number
    :param recording_data_address:\
        The address of the recording data from which to read the number
    :rtype: int
    """
    data = transceiver.read_memory(
        placement.x, placement.y,
        recording_data_address + _LAST_SEQUENCE_NUMBER_OFFSET, 4)
    return struct.unpack_from("<I", data)[0]


def get_region_pointer(placement, transceiver, recording_data_address, region):
    """ Get a pointer to a recording region

    :param placement: The placement from which to read the pointer
    :param transceiver: The transceiver to use to read the pointer
    :param recording_data_address:\
        The address of the recording data from which to read the pointer
    :param region: The index of the region to get the pointer of
    :rtype: int
    """
    data = transceiver.read_memory(
        placement.x, placement.y,
        recording_data_address + _FIRST_REGION_ADDRESS_OFFSET + (region * 4),
        4)
    return struct.unpack_from("<I", data)[0]


def get_n_timesteps_in_buffer_space(buffer_space, buffered_sdram_per_timestep):
    """ Get the number of time steps of data that can be stored in a given\
        buffers space

    :param buffer_space: The space that will hold the data
    :type buffer_space: int
    :param buffered_sdram_per_timestep:\
        The maximum SDRAM used by each region per timestep
    :type buffered_sdram_per_timestep: list of int
    :rtype: int
    """
    total_per_timestep = sum(buffered_sdram_per_timestep)
    if total_per_timestep == 0:
        return sys.maxint
    return int(math.floor(buffer_space / total_per_timestep))


def get_recorded_region_ids(buffered_sdram_per_timestep):
    """ Get the ids of regions where recording is enabled

    :param buffered_sdram_per_timestep:\
        The maximum SDRAM used by each region per timestep, where 0 indicates\
        a disabled region
    :type buffered_sdram_per_timestep: list of int
    :rtype: list of int
    """
    return [
        i for i in range(len(buffered_sdram_per_timestep))
        if buffered_sdram_per_timestep[i] > 0
    ]



# spinnman imports
from multiprocessing.pool import ThreadPool
from spinnman.connections.udp_packet_connections import EIEIOConnection
from spinnman.messages.eieio.command_messages import DatabaseConfirmation
from spinnman.messages.eieio.command_messages \
    import NotificationProtocolPauseStop, NotificationProtocolStartResume

# front end common imports
from spinn_front_end_common.utilities.constants \
    import MAX_DATABASE_PATH_LENGTH
from spinn_front_end_common.utilities.exceptions import ConfigurationException

import logging


logger = logging.getLogger(__name__)


class NotificationProtocol(object):
    """ The protocol which hand shakes with external devices about the\
        database and starting execution
    """

    def __init__(self, socket_addresses, wait_for_read_confirmation):
        self._socket_addresses = socket_addresses

        # Determines whether to wait for confirmation that the database
        # has been read before starting the simulation
        self._wait_for_read_confirmation = wait_for_read_confirmation
        self._wait_pool = ThreadPool(processes=1)
        self._data_base_message_connections = list()
        for socket_address in socket_addresses:
            self._data_base_message_connections.append(EIEIOConnection(
                local_port=socket_address.listen_port,
                remote_host=socket_address.notify_host_name,
                remote_port=socket_address.notify_port_no))

    def wait_for_confirmation(self):
        """ if asked to wait for confirmation, waits for all external systems\
            to confirm that they are configured and have read the database

        :rtype: None
        """
        logger.info("*** Awaiting for a response from an external source "
                    "to state its ready for the simulation to start ***")
        self._wait_pool.close()
        self._wait_pool.join()

    def send_start_resume_notification(self):
        """ either waits till all sources have confirmed read the database and\
            are configured, and/or just sends the start notification\
            (when the system is executing)

        :rtype: None
        """
        logger.info("*** Sending start / resume message to external sources "
                    "to state the simulation has started or resumed. ***")
        if self._wait_for_read_confirmation:
            self.wait_for_confirmation()
        eieio_command_message = NotificationProtocolStartResume()
        for connection in self._data_base_message_connections:
            try:
                connection.send_eieio_message(eieio_command_message)
            except Exception:
                logger.warning(
                    "*** Failed to send start/resume notification to"
                    " external application on"
                    " {}:{} about the simulation ***".format(
                        connection.remote_ip_address,
                        connection.remote_port))

    def send_stop_pause_notification(self):
        """ sends the pause / stop notifications when the script has either\
            finished or paused

        :rtype: None
        """
        logger.info("*** Sending pause / stop message to external sources "
                    "to state the simulation has been paused or stopped. ***")
        eieio_command_message = NotificationProtocolPauseStop()
        for connection in self._data_base_message_connections:
            try:
                connection.send_eieio_message(eieio_command_message)
            except Exception:
                logger.warning(
                    "*** Failed to send stop/pause notification to"
                    " external application on"
                    " {}:{} about the simulation ***".format(
                        connection.remote_ip_address,
                        connection.remote_port))

    # noinspection PyPep8
    def send_read_notification(self, database_path):
        """ sends notifications to all devices which have expressed an\
            interest in when the database has been written


        :param database_path: the path to the database file
        :rtype: None
        """
        if database_path is not None:
            self._wait_pool.apply_async(self._send_read_notification,
                                        args=[database_path])

    def _send_read_notification(self, database_path):
        """ sends notifications to a list of socket addresses that the\
            database has been written. Message also includes the path to the\
            database

        :param database_path: the path to the database
        :rtype: None

        """
        # noinspection PyBroadException
        if database_path is not None:
            try:
                self._sent_visualisation_confirmation = True

                # add file path to database into command message.
                number_of_chars = len(database_path)
                if number_of_chars > MAX_DATABASE_PATH_LENGTH:
                    raise ConfigurationException(
                        "The file path to the database is too large to be "
                        "transmitted via the command packet, "
                        "please set the file path manually and "
                        "set the .cfg parameter [Database] send_file_path "
                        "to False")
                eieio_command_message = DatabaseConfirmation(database_path)

                # Send command and wait for response
                logger.info(
                    "*** Notifying external sources that the database is "
                    "ready for reading ***")

                # noinspection PyBroadException
                for connection in self._data_base_message_connections:
                    try:
                        connection.send_eieio_message(eieio_command_message)
                    except Exception:
                        logger.warning(
                            "*** Failed to notify external application on"
                            " {}:{} about the database ***".format(
                                connection.remote_ip_address,
                                connection.remote_port))

                # if the system needs to wait, try receiving a packet back
                if self._wait_for_read_confirmation:
                    for connection in self._data_base_message_connections:
                        try:
                            connection.receive_eieio_message()
                            logger.info(
                                "*** Confirmation from {}:{} received,"
                                " continuing ***".format(
                                    connection.remote_ip_address,
                                    connection.remote_port))
                        except Exception:
                            logger.warning(
                                "*** Failed to receive notification from"
                                " external application on"
                                " {}:{} about the database ***".format(
                                    connection.remote_ip_address,
                                    connection.remote_port))

            except Exception:
                logger.warn("problem when sending DB notification",
                            exc_info=True)

    def close(self):
        """ Closes the thread pool
        """
        self._wait_pool.close()



from spinn_front_end_common.utilities.helpful_functions \
    import generate_unique_folder_name

# general imports
from lxml import etree
import itertools
import string


class ProvenanceXMLWriter(object):
    """ Write provenance data into XML
    """

    __slots__ = []

    VALID_CHARS = frozenset(
        "-_.() {}{}".format(string.ascii_letters, string.digits))

    def __call__(self, provenance_data_items, provenance_data_path):
        """ writes provenance in xml format

        :param provenance_data_items: data items for provenance
        :param provenance_data_path: the file path to store provenance in
        :return:  None
        """

        # Group data by the first name
        items = sorted(provenance_data_items, key=lambda item: item.names[0])
        for name, group in itertools.groupby(
                items, lambda item: item.names[0]):
            filename = "".join(
                c if c in self.VALID_CHARS else '_' for c in name)

            # generate file path for xml
            file_path = generate_unique_folder_name(
                provenance_data_path, filename, ".xml")

            # Create a root node
            root = etree.Element("provenance_data_items", name=name)

            # Keep track of sub-categories
            categories = {root: dict()}

            # Go through the items and add them
            for item in group:

                # Add the "categories" for the item (any name between the first
                # and last)
                super_element = root
                cats = categories[root]
                for cat_name in item.names[1:-1]:

                    if cat_name in cats:

                        # If there is already a category of this name under the
                        # super element, use it
                        super_element = cats[cat_name]
                    else:

                        # Otherwise, create a new category under the super
                        # element
                        super_element = etree.SubElement(
                            super_element, "provenance_data_items",
                            name=cat_name)
                        cats[cat_name] = super_element
                        categories[super_element] = dict()

                    # Get the next category for the next run of the loop
                    cats = categories[super_element]

                # Add the item
                element = etree.SubElement(
                    super_element, "provenance_data_item",
                    name=item.names[-1])
                element.text = str(item.value)

            # write xml form into file provided
            with open(file_path, "w") as writer:
                writer.write(etree.tostring(root, pretty_print=True))



from .clear_iobuf_process import ClearIOBUFProcess
from .scp_clear_iobuf_request import SCPClearIOBUFRequest
from .scp_update_runtime_request import SCPUpdateRuntimeRequest
from .update_runtime_process import UpdateRuntimeProcess

__all__ = ["ClearIOBUFProcess", "SCPClearIOBUFRequest",
           "SCPUpdateRuntimeRequest", "UpdateRuntimeProcess"]



# spinn front end common
from pacman.model.abstract_classes import AbstractHasGlobalMaxAtoms
from spinn_front_end_common.abstract_models \
    import AbstractProvidesKeyToAtomMapping, AbstractRecordable, \
    AbstractSupportsDatabaseInjection

# general imports
import logging
import os
import sqlite3
import sys

logger = logging.getLogger(__name__)


def _extract_int(x):
    return None if x is None else int(x)


class DatabaseWriter(object):
    """ The interface for the database system for main front ends.\
        Any special tables needed from a front end should be done\
        by sub classes of this interface.
    """

    __slots__ = [
        # boolean flag for when the database writer has finished
        "_done",

        # the directory of where the database is to be written
        "_database_directory",

        # the path of the database
        "_database_path",

        # the path of the initialisation SQL
        "_init_sql_path",

        # the identifier for the SpiNNaker machine
        "_machine_id",

        # the database connection itself
        "_connection",

        # Mappings used to accelerate inserts
        "_machine_to_id", "_vertex_to_id", "_edge_to_id"
    ]

    def __init__(self, database_directory):
        self._done = False
        self._database_directory = database_directory
        self._database_path = os.path.join(
            self._database_directory, "input_output_database.db")
        self._init_sql_path = os.path.join(
            os.path.dirname(__file__), "db.sql")
        self._connection = None
        self._machine_to_id = dict()
        self._vertex_to_id = dict()
        self._edge_to_id = dict()

        # delete any old database
        if os.path.isfile(self._database_path):
            os.remove(self._database_path)

        # set up checks
        self._machine_id = 0

    def __enter__(self):
        self._connection = sqlite3.connect(self._database_path)
        self.create_schema()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # @UnusedVariable
        self._connection.close()
        self._connection = None
        return False

    @staticmethod
    def auto_detect_database(machine_graph):
        """ Auto detects if there is a need to activate the database system

        :param machine_graph: the machine graph of the application\
                problem space.
        :return: a bool which represents if the database is needed
        """
        for vertex in machine_graph.vertices:
            if (isinstance(vertex, AbstractSupportsDatabaseInjection) and
                    vertex.is_in_injection_mode):
                return True
        else:
            return False

    @property
    def database_path(self):
        return self._database_path

    def __insert(self, sql, *args):
        c = self._connection.cursor()
        try:
            c.execute(sql, args)
            return c.lastrowid
        except Exception:
            logger.error("problem with insertion; argument types are %s",
                         str(map((lambda x: type(x)), args)), exc_info=True)
            raise

    def create_schema(self):
        with self._connection, open(self._init_sql_path) as f:
            sql = f.read()
            self._connection.executescript(sql)

    def __insert_machine_layout(self, x_dimension, y_dimension):
        return self.__insert(
            "INSERT INTO Machine_layout("
            "  x_dimension, y_dimension) "
            "VALUES(?, ?)",
            int(x_dimension), int(y_dimension))

    def __insert_machine_chip(self, no_processors, chip, machine_id):
        return self.__insert(
            "INSERT INTO Machine_chip("
            "  no_processors, chip_x, chip_y, machine_id) "
            "VALUES (?, ?, ?, ?)",
            int(no_processors), int(chip.x), int(chip.y), int(machine_id))

    def __insert_processor(self, chip, machine_id, available_DTCM,
                           available_CPU, physical_id):
        return self.__insert(
            "INSERT INTO Processor("
            "  chip_x, chip_y, machine_id, available_DTCM, "
            "  available_CPU, physical_id) "
            "VALUES(?, ?, ?, ?, ?, ?)",
            int(chip.x), int(chip.y), int(machine_id),
            int(available_DTCM), int(available_CPU), int(physical_id))

    def __insert_app_vertex(self, vertex, max_atoms, is_recording):
        v_id = self.__insert(
            "INSERT INTO Application_vertices("
            "  vertex_label, no_atoms, max_atom_constrant, recorded) "
            "VALUES(?, ?, ?, ?)",
            str(vertex.label), int(vertex.n_atoms), int(max_atoms),
            int(is_recording))
        self._vertex_to_id[vertex] = v_id
        return v_id

    def __insert_app_edge(self, edge):
        e_id = self.__insert(
            "INSERT INTO Application_edges ("
            "  pre_vertex, post_vertex, edge_label) "
            "VALUES(?, ?, ?)",
            int(self._vertex_to_id[edge.pre_vertex]),
            int(self._vertex_to_id[edge.post_vertex]),
            str(edge.label))
        self._edge_to_id[edge] = e_id
        return e_id

    def __insert_app_graph_element(self, vertex, edge):
        return self.__insert(
            "INSERT INTO Application_graph ("
            "  vertex_id, edge_id) "
            "VALUES(?, ?)",
            int(self._vertex_to_id[vertex]),
            int(self._edge_to_id[edge]))

    def __insert_cfg(self, parameter_id, value):
        # NB: No type constraints on value; this is SQLite (not Sparta!)
        return self.__insert(
            "INSERT INTO configuration_parameters ("
            "  parameter_id, value) "
            "VALUES (?, ?)",
            str(parameter_id), value)

    def __insert_machine_vertex(self, vertex, cpu_used, sdram_used, dtcm_used):
        v_id = self.__insert(
            "INSERT INTO Machine_vertices ("
            "  label, cpu_used, sdram_used, dtcm_used) "
            "VALUES(?, ?, ?, ?)",
            str(vertex.label), _extract_int(cpu_used.get_value()),
            _extract_int(sdram_used.get_value()),
            _extract_int(dtcm_used.get_value()))
        self._vertex_to_id[vertex] = v_id
        return v_id

    def __insert_machine_edge(self, edge):
        e_id = self.__insert(
            "INSERT INTO Machine_edges ("
            "  pre_vertex, post_vertex, label) "
            "VALUES(?, ?, ?)",
            int(self._vertex_to_id[edge.pre_vertex]),
            int(self._vertex_to_id[edge.post_vertex]),
            str(edge.label))
        self._edge_to_id[edge] = e_id
        return e_id

    def __insert_machine_graph_element(self, vertex, edge):
        return self.__insert(
            "INSERT INTO Machine_graph ("
            "  vertex_id, edge_id) "
            "VALUES(?, ?)",
            int(self._vertex_to_id[vertex]), int(self._edge_to_id[edge]))

    def __insert_graph_mapper_vertex(
            self, app_vertex, machine_vertex, vertex_slice):
        return self.__insert(
            "INSERT INTO graph_mapper_vertex ("
            "  application_vertex_id, machine_vertex_id, "
            "  lo_atom, hi_atom) "
            "VALUES(?, ?, ?, ?)",
            int(self._vertex_to_id[app_vertex]),
            int(self._vertex_to_id[machine_vertex]),
            int(vertex_slice.lo_atom), int(vertex_slice.hi_atom))

    def __insert_graph_mapper_edge(self, app_edge, machine_edge):
        return self.__insert(
            "INSERT INTO graph_mapper_edges ("
            "  application_edge_id, machine_edge_id) "
            "VALUES(?, ?)",
            int(self._edge_to_id[app_edge]),
            int(self._edge_to_id[machine_edge]))

    def __insert_placement(self, placement, machine_id):
        return self.__insert(
            "INSERT INTO Placements("
            "  vertex_id, chip_x, chip_y, chip_p, machine_id) "
            "VALUES(?, ?, ?, ?, ?)",
            int(self._vertex_to_id[placement.vertex]),
            int(placement.x), int(placement.y), int(placement.p),
            int(machine_id))

    def __insert_routing_info(self, edge, key_mask):
        return self.__insert(
            "INSERT INTO Routing_info("
            "  edge_id, \"key\", mask) "
            "VALUES(?, ?, ?)",
            int(self._edge_to_id[edge]), int(key_mask.key), int(key_mask.mask))

    def __insert_routing_entry(self, routing_table, counter, entry, route):
        return self.__insert(
            "INSERT INTO Routing_table("
            "  chip_x, chip_y, position, key_combo, mask, route) "
            "VALUES(?, ?, ?, ?, ?, ?)",
            int(routing_table.x), int(routing_table.y), int(counter),
            int(entry.routing_entry_key), int(entry.mask), int(route))

    def __insert_ip_tag(self, vertex, ip_tag):
        return self.__insert(
            "INSERT INTO IP_tags("
            "  vertex_id, tag, board_address, ip_address,"
            "  port, strip_sdp) "
            "VALUES (?, ?, ?, ?, ?, ?)",
            int(self._vertex_to_id[vertex]),
            int(ip_tag.tag), str(ip_tag.board_address),
            str(ip_tag.ip_address), int(ip_tag.port),
            1 if ip_tag.strip_sdp else 0)

    def __insert_reverse_ip_tag(self, vertex, reverse_ip_tag):
        return self.__insert(
            "INSERT INTO Reverse_IP_tags("
            "  vertex_id, tag, board_address, port) "
            "VALUES (?, ?, ?, ?)",
            int(self._vertex_to_id[vertex]), int(reverse_ip_tag.tag),
            str(reverse_ip_tag.board_address), int(reverse_ip_tag.port))

    def __insert_event_atom_mapping(self, vertex, event_id, atom_id):
        return self.__insert(
            "INSERT INTO event_to_atom_mapping("
            "  vertex_id, event_id, atom_id) "
            "VALUES (?, ?, ?)",
            int(self._vertex_to_id[vertex]), int(event_id), int(atom_id))

    def add_machine_objects(self, machine):
        """ Store the machine object into the database

        :param machine: the machine object.
        :rtype: None
        """
        with self._connection:
            self._machine_to_id[machine] = self.__insert_machine_layout(
                machine.max_chip_x + 1, machine.max_chip_y + 1)
            self._machine_id += 1
            for chip in machine.chips:
                self.__insert_machine_chip(
                    len(list(chip.processors)), chip, self._machine_id)
                for processor in chip.processors:
                    self.__insert_processor(
                        chip, self._machine_id,
                        processor.dtcm_available,
                        processor.cpu_cycles_available,
                        processor.processor_id)

    def add_application_vertices(self, application_graph):
        """

        :param application_graph:
        :rtype: None
        """
        with self._connection:
            # add vertices
            for vertex in application_graph.vertices:
                if isinstance(vertex, AbstractRecordable):
                    self.__insert_app_vertex(
                        vertex, vertex.get_max_atoms_per_core(),
                        vertex.is_recording_spikes())
                elif isinstance(vertex, AbstractHasGlobalMaxAtoms):
                    self.__insert_app_vertex(
                        vertex, vertex.get_max_atoms_per_core(), 0)
                else:
                    self.__insert_app_vertex(vertex, sys.maxint, 0)

            # add edges
            for vertex in application_graph.vertices:
                for edge in application_graph.\
                        get_edges_starting_at_vertex(vertex):
                    self.__insert_app_edge(edge)

            # update graph
            for vertex in application_graph.vertices:
                for edge in application_graph.\
                        get_edges_starting_at_vertex(vertex):
                    self.__insert_app_graph_element(vertex, edge)

    def add_system_params(self, time_scale_factor, machine_time_step, runtime):
        """ Write system params into the database

        :param time_scale_factor: the time scale factor used in timing
        :param machine_time_step: the machine time step used in timing
        :param runtime: the amount of time the application is to run for
        """
        with self._connection:
            self.__insert_cfg("machine_time_step", machine_time_step)
            self.__insert_cfg("time_scale_factor", time_scale_factor)
            if runtime is not None:
                self.__insert_cfg("infinite_run", "False")
                self.__insert_cfg("runtime", runtime)
            else:
                self.__insert_cfg("infinite_run", "True")
                self.__insert_cfg("runtime", -1)

    def add_vertices(self, machine_graph, graph_mapper, application_graph):
        """ Add the machine graph, graph mapper and application graph \
            into the database.

        :param machine_graph: the machine graph object
        :param graph_mapper: the graph mapper object
        :param application_graph: the application graph object
        :rtype: None
        """
        with self._connection:
            for vertex in machine_graph.vertices:
                req = vertex.resources_required
                self.__insert_machine_vertex(
                    vertex, req.cpu_cycles, req.sdram, req.dtcm)

            # add machine edges
            for edge in machine_graph.edges:
                self.__insert_machine_edge(edge)

            # add to machine graph
            for vertex in machine_graph.vertices:
                for edge in machine_graph.get_edges_starting_at_vertex(vertex):
                    self.__insert_machine_graph_element(vertex, edge)

            if application_graph is not None:
                for machine_vertex in machine_graph.vertices:
                    app_vertex = graph_mapper.get_application_vertex(
                        machine_vertex)
                    self.__insert_graph_mapper_vertex(
                        app_vertex, machine_vertex,
                        graph_mapper.get_slice(machine_vertex))

                # add graph_mapper edges
                for edge in machine_graph.edges:
                    self.__insert_graph_mapper_edge(
                        graph_mapper.get_application_edge(edge), edge)

    def add_placements(self, placements):
        """ Adds the placements objects into the database

        :param placements: the placements object
        :param machine_graph: the machine graph object
        :rtype: None
        """
        with self._connection:
            # add records
            for placement in placements.placements:
                self.__insert_placement(placement, self._machine_id)

    def add_routing_infos(self, routing_infos, machine_graph):
        """ Adds the routing information (key masks etc) into the database

        :param routing_infos: the routing information object
        :param machine_graph: the machine graph object
        :rtype: None:
        """
        with self._connection:
            for partition in machine_graph.outgoing_edge_partitions:
                rinfo = routing_infos.get_routing_info_from_partition(
                    partition)
                for edge in partition.edges:
                    for key_mask in rinfo.keys_and_masks:
                        self.__insert_routing_info(edge, key_mask)

    def add_routing_tables(self, routing_tables):
        """ Adds the routing tables into the database

        :param routing_tables: the routing tables object
        :rtype: None
        """
        with self._connection:
            for routing_table in routing_tables.routing_tables:
                for counter, entry in \
                        enumerate(routing_table.multicast_routing_entries):
                    route_entry = 0
                    for processor_id in entry.processor_ids:
                        route_entry |= 1 << (6 + processor_id)
                    for link_id in entry.link_ids:
                        route_entry |= 1 << link_id
                    self.__insert_routing_entry(
                        routing_table, counter, entry, route_entry)

    def add_tags(self, machine_graph, tags):
        """ Adds the tags into the database

        :param machine_graph: the machine graph object
        :param tags: the tags object
        :rtype: None
        """
        with self._connection:
            for vertex in machine_graph.vertices:
                ip_tags = tags.get_ip_tags_for_vertex(vertex)
                if ip_tags is not None:
                    for ip_tag in ip_tags:
                        self.__insert_ip_tag(vertex, ip_tag)
                reverse_ip_tags = tags.get_reverse_ip_tags_for_vertex(
                    vertex)
                if reverse_ip_tags is not None:
                    for reverse_ip_tag in reverse_ip_tags:
                        self.__insert_reverse_ip_tag(vertex, reverse_ip_tag)

    def create_atom_to_event_id_mapping(
            self, application_graph, machine_graph, routing_infos,
            graph_mapper):
        """

        :param application_graph:
        :param machine_graph:
        :param routing_infos:
        :param graph_mapper:
        :rtype: None
        """
        have_app_graph = (application_graph is not None and
                          application_graph.n_vertices != 0)
        with self._connection:
            for vertex in machine_graph.vertices:
                for partition in machine_graph.\
                        get_outgoing_edge_partitions_starting_at_vertex(
                            vertex):
                    if have_app_graph:
                        self._insert_vertex_atom_to_key_map(
                            graph_mapper.get_application_vertex(vertex),
                            partition, routing_infos)
                    else:
                        self._insert_vertex_atom_to_key_map(
                            vertex, partition, routing_infos)

    def _insert_vertex_atom_to_key_map(
            self, vertex, partition, routing_infos):
        """

        :param vertex:
        :param partition:
        :param routing_infos:
        :rtype: None
        """
        if isinstance(vertex, AbstractProvidesKeyToAtomMapping):
            routing_info = routing_infos.get_routing_info_from_partition(
                partition)
            for atom_id, key in vertex.routing_key_partition_atom_mapping(
                    routing_info, partition):
                self.__insert_event_atom_mapping(vertex, key, atom_id)



from .buffered_receiving_data import BufferedReceivingData
from .buffered_sending_region import BufferedSendingRegion
from .buffers_sent_deque import BuffersSentDeque
from .channel_buffer_state import ChannelBufferState
from .end_buffering_state import EndBufferingState

__all__ = ["BufferedReceivingData", "BufferedSendingRegion",
           "BuffersSentDeque", "ChannelBufferState", "EndBufferingState"]



from enum import Enum

LIVE_GATHERER_CORE_APPLICATION_ID = 0xAC0
COMMAND_SENDER_CORE_APPLICATION_ID = 0xAC6
SPIKE_INJECTOR_CORE_APPLICATION_ID = 0xAC9

BITS_PER_WORD = 32.0
SDRAM_BASE_ADDR = 0x70000000
MAX_SAFE_BINARY_SIZE = 32 * 1024
MAX_POSSIBLE_BINARY_SIZE = 33 * 1024

# max size expected to be used by the reverse ip_tag multicast source
# during buffered operations
MAX_SIZE_OF_BUFFERED_REGION_ON_CHIP = 1 * 1024 * 1024

# The default size of a recording buffer before receive request is sent
DEFAULT_BUFFER_SIZE_BEFORE_RECEIVE = 16 * 1024

# The number of words in the AbstractDataSpecable basic setup information
# 4 for MAGIC NUMBER
# 4 for DSG VERSION
DATA_SPECABLE_BASIC_SETUP_INFO_N_BYTES = 8

# The number of words used by the simulation interface
# 4 for machine_time_step,
# 4 for SDP port
# 4 for application hash
SIMULATION_N_BYTES = 12

# The number of bytes used by the dsg and simulation interfaces
SYSTEM_BYTES_REQUIREMENT = \
    (DATA_SPECABLE_BASIC_SETUP_INFO_N_BYTES + SIMULATION_N_BYTES)

# The number of bytes used by SARK per memory allocation
SARK_PER_MALLOC_SDRAM_USAGE = 8

# database cap file path
MAX_DATABASE_PATH_LENGTH = 50000

# size of the on-chip DSE data structure required in bytes
DSE_DATA_STRUCT_SIZE = 16

SDP_RUNNING_MESSAGE_CODES = Enum(
    value="SDP_RUNNING_MESSAGE_ID_CODES",
    names=[
        ("SDP_STOP_ID_CODE", 6),
        ("SDP_NEW_RUNTIME_ID_CODE", 7),
        ("SDP_UPDATE_PROVENCE_REGION_AND_EXIT", 8),
        ("SDP_CLEAR_IOBUF_CODE", 9)])


# SDP port handling output buffering data streaming
SDP_PORTS = Enum(
    value="SDP_PORTS",
    names=[

        # command port for the buffered in functionality
        ("INPUT_BUFFERING_SDP_PORT", 1),

        # command port for the buffered out functionality
        ("OUTPUT_BUFFERING_SDP_PORT", 2),

        # command port for resetting runtime etc
        ("RUNNING_COMMAND_SDP_PORT", 3)]
)

# output buffering operations
# a listing of what spinnaker specific EIEIO commands there are.
BUFFERING_OPERATIONS = Enum(
    value="BUFFERING_OPERATIONS",
    names=[

        # Database handshake with external program
        ("BUFFER_READ", 0),

        # Host confirming data being read form SpiNNaker memory
        ("BUFFER_WRITE", 1)]
)



# pacman imports
from pacman.model.decorators import overrides
from pacman.model.graphs.application import ApplicationVertex
from pacman.model.resources import CPUCyclesPerTickResource, DTCMResource
from pacman.model.resources import ResourceContainer, SDRAMResource
from pacman.model.resources import ReverseIPtagResource, IPtagResource
from pacman.model.constraints.placer_constraints import BoardConstraint


# front end common imports
from spinn_front_end_common.abstract_models \
    import AbstractProvidesOutgoingPartitionConstraints
from spinn_front_end_common.abstract_models.impl\
    import ProvidesKeyToAtomMappingImpl
from spinn_front_end_common.utilities import constants
from .reverse_ip_tag_multicast_source_machine_vertex \
    import ReverseIPTagMulticastSourceMachineVertex
from spinn_front_end_common.abstract_models \
    import AbstractGeneratesDataSpecification, AbstractHasAssociatedBinary
from spinn_front_end_common.interface.buffer_management \
    import recording_utilities
from spinn_front_end_common.utilities.utility_objs import ExecutableStartType

# general imports
import sys


class ReverseIpTagMultiCastSource(
        ApplicationVertex, AbstractGeneratesDataSpecification,
        AbstractHasAssociatedBinary,
        AbstractProvidesOutgoingPartitionConstraints,
        ProvidesKeyToAtomMappingImpl):
    """ A model which will allow events to be injected into a spinnaker\
        machine and converted into multicast packets.
    """

    def __init__(
            self, n_keys, label=None, constraints=None,
            max_atoms_per_core=sys.maxint,

            # General parameters
            board_address=None,

            # Live input parameters
            receive_port=None,
            receive_sdp_port=(
                constants.SDP_PORTS.INPUT_BUFFERING_SDP_PORT.value),
            receive_tag=None,
            receive_rate=10,

            # Key parameters
            virtual_key=None, prefix=None,
            prefix_type=None, check_keys=False,

            # Send buffer parameters
            send_buffer_times=None,
            send_buffer_partition_id=None,
            send_buffer_max_space=(
                constants.MAX_SIZE_OF_BUFFERED_REGION_ON_CHIP),
            send_buffer_space_before_notify=640,

            # Buffer parameters
            buffer_notification_ip_address=None,
            buffer_notification_port=None,
            buffer_notification_tag=None,

            # Extra flag for input without a reserved port
            reserve_reverse_ip_tag=False):
        """

        :param n_keys: The number of keys to be sent via this multicast source
        :param label: The label of this vertex
        :param constraints: Any initial constraints to this vertex
        :param board_address: The IP address of the board on which to place\
                this vertex if receiving data, either buffered or live (by\
                default, any board is chosen)
        :param receive_port: The port on the board that will listen for\
                incoming event packets (default is to disable this feature;\
                set a value to enable it)
        :param receive_sdp_port: The SDP port to listen on for incoming event\
                packets (defaults to 1)
        :param receive_tag: The IP tag to use for receiving live events\
                (uses any by default)
        :param receive_rate: The estimated rate of packets that will be sent\
                by this source
        :param virtual_key: The base multicast key to send received events\
                with (assigned automatically by default)
        :param prefix: The prefix to "or" with generated multicast keys\
                (default is no prefix)
        :param prefix_type: Whether the prefix should apply to the upper or\
                lower half of the multicast keys (default is upper half)
        :param check_keys: True if the keys of received events should be\
                verified before sending (default False)
        :param send_buffer_times: An array of arrays of times at which keys\
                should be sent (one array for each key, default disabled)
        :param send_buffer_partition_id: The id of the partition containing\
                the edges down which the events are to be sent
        :param send_buffer_max_space: The maximum amount of space to use of\
                the SDRAM on the machine (default is 1MB)
        :param send_buffer_space_before_notify: The amount of space free in\
                the sending buffer before the machine will ask the host for\
                more data (default setting is optimised for most cases)
        :param buffer_notification_ip_address: The IP address of the host\
                that will send new buffers (must be specified if a send buffer\
                is specified or if recording will be used)
        :param buffer_notification_port: The port that the host that will\
                send new buffers is listening on (must be specified if a\
                send buffer is specified, or if recording will be used)
        :param buffer_notification_tag: The IP tag to use to notify the\
                host about space in the buffer (default is to use any tag)
        """
        ApplicationVertex.__init__(
            self, label, constraints, max_atoms_per_core)
        ProvidesKeyToAtomMappingImpl.__init__(self)

        # basic items
        self._n_atoms = n_keys

        # Store the parameters for EIEIO
        self._board_address = board_address
        self._receive_port = receive_port
        self._receive_sdp_port = receive_sdp_port
        self._receive_tag = receive_tag
        self._receive_rate = receive_rate
        self._virtual_key = virtual_key
        self._prefix = prefix
        self._prefix_type = prefix_type
        self._check_keys = check_keys

        self._reverse_iptags = None
        if receive_port is not None or reserve_reverse_ip_tag:
            self._reverse_iptags = [ReverseIPtagResource(
                port=receive_port, sdp_port=receive_sdp_port,
                tag=receive_tag)]
            if board_address is not None:
                self.add_constraint(BoardConstraint(board_address))

        # Store the send buffering details
        self._send_buffer_times = send_buffer_times
        self._send_buffer_partition_id = send_buffer_partition_id
        self._send_buffer_max_space = send_buffer_max_space
        self._send_buffer_space_before_notify = send_buffer_space_before_notify

        # Store the buffering details
        self._buffer_notification_ip_address = buffer_notification_ip_address
        self._buffer_notification_port = buffer_notification_port
        self._buffer_notification_tag = buffer_notification_tag
        self._reserve_reverse_ip_tag = reserve_reverse_ip_tag

        self._iptags = None
        if send_buffer_times is not None:
            self._iptags = [IPtagResource(
                buffer_notification_ip_address, buffer_notification_port, True,
                buffer_notification_tag)]
            if board_address is not None:
                self.add_constraint(BoardConstraint(board_address))

        # Store recording parameters
        self._record_buffer_size = 0
        self._record_buffer_size_before_receive = 0
        self._record_time_between_requests = 0

        # Keep the vertices for resuming runs
        self._machine_vertices = list()

    @property
    @overrides(ApplicationVertex.n_atoms)
    def n_atoms(self):
        return self._n_atoms

    @overrides(ApplicationVertex.get_resources_used_by_atoms)
    def get_resources_used_by_atoms(self, vertex_slice):
        container = ResourceContainer(
            sdram=SDRAMResource(
                ReverseIPTagMulticastSourceMachineVertex.get_sdram_usage(
                    self._send_buffer_times, self._send_buffer_max_space,
                    self._record_buffer_size > 0)),
            dtcm=DTCMResource(
                ReverseIPTagMulticastSourceMachineVertex.get_dtcm_usage()),
            cpu_cycles=CPUCyclesPerTickResource(
                ReverseIPTagMulticastSourceMachineVertex.get_cpu_usage()),
            iptags=self._iptags,
            reverse_iptags=self._reverse_iptags)
        if self._iptags is None:
            container.extend(recording_utilities.get_recording_resources(
                [self._record_buffer_size],
                self._buffer_notification_ip_address,
                self._buffer_notification_port, self._buffer_notification_tag))
        else:
            container.extend(recording_utilities.get_recording_resources(
                [self._record_buffer_size]))
        return container

    @property
    def send_buffer_times(self):
        return self._send_buffer_times

    @send_buffer_times.setter
    def send_buffer_times(self, send_buffer_times):
        self._send_buffer_times = send_buffer_times
        for (vertex_slice, vertex) in self._machine_vertices:
            send_buffer_times_to_set = self._send_buffer_times
            if (self._send_buffer_times is not None and
                    len(self._send_buffer_times) > 0):
                if hasattr(self._send_buffer_times[0], "__len__"):
                    send_buffer_times_to_set = self._send_buffer_times[
                        vertex_slice.lo_atom:vertex_slice.hi_atom + 1]
            vertex.send_buffer_times = send_buffer_times_to_set

    def enable_recording(
            self,
            record_buffer_size=constants.MAX_SIZE_OF_BUFFERED_REGION_ON_CHIP,
            buffer_size_before_receive=(
                constants.DEFAULT_BUFFER_SIZE_BEFORE_RECEIVE),
            time_between_requests=0):
        self._record_buffer_size = record_buffer_size
        self._record_buffer_size_before_receive = buffer_size_before_receive
        self._record_time_between_requests = time_between_requests

    @overrides(AbstractProvidesOutgoingPartitionConstraints.
               get_outgoing_partition_constraints)
    def get_outgoing_partition_constraints(self, partition):
        return partition.pre_vertex.get_outgoing_partition_constraints(
            partition)

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self):
        return 'reverse_iptag_multicast_source.aplx'

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self):
        return ExecutableStartType.USES_SIMULATION_INTERFACE

    def generate_data_specification(self, spec, placement):
        placement.vertex.generate_data_specification(spec, placement)

    @overrides(ApplicationVertex.create_machine_vertex)
    def create_machine_vertex(
            self, vertex_slice, resources_required, label=None,
            constraints=None):
        send_buffer_times = self._send_buffer_times
        if (self._send_buffer_times is not None and
                len(self._send_buffer_times) > 0):
            if hasattr(self._send_buffer_times[0], "__len__"):
                send_buffer_times = self._send_buffer_times[
                    vertex_slice.lo_atom:vertex_slice.hi_atom + 1]
        vertex = ReverseIPTagMulticastSourceMachineVertex(
            n_keys=vertex_slice.n_atoms,
            label=label, constraints=constraints,
            board_address=self._board_address,
            receive_port=self._receive_port,
            receive_sdp_port=self._receive_sdp_port,
            receive_tag=self._receive_tag,
            receive_rate=self._receive_rate,
            virtual_key=self._virtual_key, prefix=self._prefix,
            prefix_type=self._prefix_type, check_keys=self._check_keys,
            send_buffer_times=send_buffer_times,
            send_buffer_partition_id=self._send_buffer_partition_id,
            send_buffer_max_space=self._send_buffer_max_space,
            send_buffer_space_before_notify=(
                self._send_buffer_space_before_notify),
            buffer_notification_ip_address=(
                self._buffer_notification_ip_address),
            buffer_notification_port=self._buffer_notification_port,
            buffer_notification_tag=self._buffer_notification_tag,
            reserve_reverse_ip_tag=self._reserve_reverse_ip_tag)
        if self._record_buffer_size > 0:
            vertex.enable_recording(
                self._record_buffer_size,
                self._record_buffer_size_before_receive,
                self._record_time_between_requests)
        self._machine_vertices.append((vertex_slice, vertex))
        return vertex

    def __repr__(self):
        return self._label



from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.interface.buffer_management.buffer_models \
    import AbstractReceiveBuffersToHost
from spinn_utilities.progress_bar import ProgressBar


class BufferExtractor(object):
    """ Extracts data in between runs
    """

    __slots__ = []

    def __call__(self, machine_graph, placements, buffer_manager, ran_token):
        if not ran_token:
            raise ConfigurationException("The ran token has not been set")

        # Count the regions to be read
        n_regions_to_read = 0
        vertices = list()
        for vertex in machine_graph.vertices:
            if isinstance(vertex, AbstractReceiveBuffersToHost):
                n_regions_to_read += len(vertex.get_recorded_region_ids())
                vertices.append(vertex)

        progress = ProgressBar(
            n_regions_to_read, "Extracting buffers from the last run")

        # Read back the regions
        for vertex in vertices:
            placement = placements.get_placement_of_vertex(vertex)
            for recording_region_id in vertex.get_recorded_region_ids():
                buffer_manager.get_data_for_vertex(
                    placement, recording_region_id)
                progress.update()
        progress.end()



from spinn_front_end_common.abstract_models.\
    abstract_has_associated_binary import AbstractHasAssociatedBinary
from spinn_front_end_common.utilities.utility_objs.executable_start_type \
    import ExecutableStartType
from spinn_front_end_common.utilities import exceptions
from spinn_utilities.progress_bar import ProgressBar


class LocateExecutableStartType(object):

    def __call__(self, graph, graph_mapper=None):

        progress = ProgressBar(
            graph.n_vertices, "Finding executable_start_types")
        binary_start_type = None
        if len(graph.vertices) == 0:
            return ExecutableStartType.NO_APPLICATION
        for vertex in progress.over(graph.vertices):

            placement_binary_start_type = None

            if isinstance(vertex, AbstractHasAssociatedBinary):
                placement_binary_start_type = vertex.get_binary_start_type()
            elif graph_mapper is not None:
                associated_vertex = graph_mapper.get_application_vertex(vertex)
                if isinstance(associated_vertex, AbstractHasAssociatedBinary):
                    placement_binary_start_type = \
                        associated_vertex.get_binary_start_type()

            if binary_start_type is None:
                binary_start_type = placement_binary_start_type

            # check all vertices have the same start type
            if (placement_binary_start_type is not None and
                    binary_start_type is not None and
                    placement_binary_start_type != binary_start_type):
                raise exceptions.ConfigurationException(
                    "All binaries must be of the same start type -"
                    " existing binaries have start type {} but "
                    "vertex {} has start type {}".format(
                        binary_start_type, vertex.label,
                        placement_binary_start_type))

        if binary_start_type is None:
            return ExecutableStartType.NO_APPLICATION

        return binary_start_type



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase, abstractproperty


@add_metaclass(AbstractBase)
class AbstractSupportsDatabaseInjection(object):
    """Marks a machine vertex as supporting injection of information via a \
    database running on the controlling host.
    """

    __slots__ = ()

    @abstractproperty
    def is_in_injection_mode(self):
        """Whether this vertex is actually in injection mode."""




import logging
import os

logger = logging.getLogger(__name__)


class MemoryMapOnHostReport(object):
    """ Report on memory usage
    """

    def __call__(
            self, report_default_directory,
            processor_to_app_data_base_address):
        """

        :param report_default_directory:
        :param processor_to_app_data_base_address:
        :rtype: None
        """

        file_name = os.path.join(report_default_directory,
                                 "memory_map_from_processor_to_address_space")
        output = None
        try:
            output = open(file_name, "w")
        except IOError:
            logger.error("Generate_placement_reports: Can't open file"
                         " {} for writing.".format(file_name))

        output.write("On host data specification executor\n")

        for key in processor_to_app_data_base_address:
            data = processor_to_app_data_base_address[key]
            output.write(
                "{}: ('start_address': {}, hex({}), 'memory_used': {}, "
                "'memory_written': {} \n".format(
                    key, data['start_address'], hex(data['start_address']),
                    data['memory_used'], data['memory_written']))

        output.flush()
        output.close()



from enum import Enum
from collections import Counter

from pacman.model.decorators import overrides
from pacman.model.graphs.machine import MachineVertex
from spinn_front_end_common.abstract_models import AbstractHasAssociatedBinary
from spinn_front_end_common.interface.provenance \
    import ProvidesProvenanceDataFromMachineImpl
from spinn_front_end_common.interface.simulation.simulation_utilities \
    import get_simulation_header_array
from spinn_front_end_common.utilities.constants \
    import SYSTEM_BYTES_REQUIREMENT
from spinn_front_end_common.utilities.utility_objs import ExecutableStartType


class CommandSenderMachineVertex(
        MachineVertex, ProvidesProvenanceDataFromMachineImpl,
        AbstractHasAssociatedBinary):

    # Regions for populations
    DATA_REGIONS = Enum(
        value="DATA_REGIONS",
        names=[('SYSTEM_REGION', 0),
               ('SETUP', 1),
               ('COMMANDS_WITH_ARBITRARY_TIMES', 2),
               ('COMMANDS_AT_START_RESUME', 3),
               ('COMMANDS_AT_STOP_PAUSE', 4),
               ('PROVENANCE_REGION', 5)])

    # 4 for key, 4 for has payload, 4 for payload 4 for repeats, 4 for delays
    _COMMAND_WITH_PAYLOAD_SIZE = 20

    # 4 for the time stamp
    _COMMAND_TIMESTAMP_SIZE = 4

    # 4 for the int to represent the number of commands
    _N_COMMANDS_SIZE = 4

    # bool for if the command has a payload (true = 1)
    _HAS_PAYLOAD = 1

    # bool for if the command does not have a payload (false = 0)
    _HAS_NO_PAYLOAD = 0

    # Setup data size (one word)
    _SETUP_DATA_SIZE = 4

    # the number of malloc requests used by the dsg
    TOTAL_REQUIRED_MALLOCS = 5

    # The name of the binary file
    BINARY_FILE_NAME = 'command_sender_multicast_source.aplx'

    def __init__(
            self, constraints, resources_required, label,
            commands_at_start_resume, commands_at_pause_stop, timed_commands):
        ProvidesProvenanceDataFromMachineImpl.__init__(self)
        MachineVertex.__init__(self, label, constraints)

        # container of different types of command
        self._timed_commands = timed_commands
        self._commands_at_start_resume = commands_at_start_resume
        self._commands_at_pause_stop = commands_at_pause_stop
        self._resources = resources_required

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._provenance_region_id)
    def _provenance_region_id(self):
        return self.DATA_REGIONS.PROVENANCE_REGION.value

    @property
    @overrides(ProvidesProvenanceDataFromMachineImpl._n_additional_data_items)
    def _n_additional_data_items(self):
        return 0

    @property
    @overrides(MachineVertex.resources_required)
    def resources_required(self):
        return self._resources

    def generate_data_specification(
            self, spec, placement, machine_time_step, time_scale_factor,
            n_machine_time_steps):

        timed_commands_size = \
            self.get_timed_commands_bytes(self._timed_commands)
        start_resume_commands_size = \
            self.get_n_command_bytes(self._commands_at_start_resume)
        pause_stop_commands_size = \
            self.get_n_command_bytes(self._commands_at_pause_stop)

        # reverse memory regions
        self._reserve_memory_regions(
            spec, timed_commands_size, start_resume_commands_size,
            pause_stop_commands_size, placement.vertex)

        # Write system region
        spec.comment("\n*** Spec for multicast source ***\n\n")
        spec.switch_write_focus(
            CommandSenderMachineVertex.DATA_REGIONS.SYSTEM_REGION.value)
        spec.write_array(get_simulation_header_array(
            self.get_binary_file_name(), machine_time_step,
            time_scale_factor))

        # Write setup region
        # Find the maximum number of commands per timestep
        max_n_commands = 0
        if len(self._timed_commands) > 0:
            counter = Counter(self._timed_commands)
            max_n_commands = counter.most_common(1)[0][1]
        max_n_commands = max([
            max_n_commands, len(self._commands_at_start_resume),
            len(self._commands_at_pause_stop)])
        time_between_commands = 0
        if max_n_commands > 0:
            time_between_commands = (
                (machine_time_step * time_scale_factor / 2) / max_n_commands)
        spec.switch_write_focus(
            CommandSenderMachineVertex.DATA_REGIONS.SETUP.value)
        spec.write_value(time_between_commands)

        # write commands
        spec.switch_write_focus(
            region=CommandSenderMachineVertex.DATA_REGIONS.
            COMMANDS_WITH_ARBITRARY_TIMES.value)

        # write commands to spec for timed commands
        self._write_timed_commands(self._timed_commands, spec)

        # write commands fired off during a start or resume
        spec.switch_write_focus(
            region=CommandSenderMachineVertex.DATA_REGIONS.
            COMMANDS_AT_START_RESUME.value)

        self._write_basic_commands(self._commands_at_start_resume, spec)

        # write commands fired off during a pause or end
        spec.switch_write_focus(
            region=CommandSenderMachineVertex.DATA_REGIONS.
            COMMANDS_AT_STOP_PAUSE.value)

        self._write_basic_commands(self._commands_at_pause_stop, spec)

        # End-of-Spec:
        spec.end_specification()

    def _write_basic_commands(self, commands, spec):

        # number of commands
        spec.write_value(len(commands))

        # write commands to region
        for command in commands:
            self._write_command(command, spec)

    def _write_timed_commands(self, timed_commands, spec):

        spec.write_value(len(timed_commands))

        # write commands
        for command in timed_commands:
            spec.write_value(command.time)
            self._write_command(command, spec)

    @staticmethod
    def _write_command(command, spec):
        spec.write_value(command.key)
        spec.write_value(CommandSenderMachineVertex._HAS_PAYLOAD)
        spec.write_value(command.payload if command.is_payload else 0)
        spec.write_value(command.repeat)
        spec.write_value(command.delay_between_repeats)

    @staticmethod
    def _reserve_memory_regions(
            spec, time_command_size, start_command_size, end_command_size,
            vertex):
        """
        Reserve SDRAM space for memory areas:
        1) Area for information on what data to record
        2) area for start commands
        3) area for end commands
        """
        spec.comment("\nReserving memory space for data regions:\n\n")

        # Reserve memory:
        spec.reserve_memory_region(
            region=CommandSenderMachineVertex.DATA_REGIONS.SYSTEM_REGION.value,
            size=SYSTEM_BYTES_REQUIREMENT, label='system')

        spec.reserve_memory_region(
            region=CommandSenderMachineVertex.DATA_REGIONS.SETUP.value,
            size=CommandSenderMachineVertex._SETUP_DATA_SIZE, label='setup')

        spec.reserve_memory_region(
            region=CommandSenderMachineVertex.
            DATA_REGIONS.COMMANDS_WITH_ARBITRARY_TIMES.value,
            size=time_command_size, label='commands with arbitrary times')

        spec.reserve_memory_region(
            region=CommandSenderMachineVertex.
            DATA_REGIONS.COMMANDS_AT_START_RESUME.value,
            size=start_command_size, label='commands with start resume times')

        spec.reserve_memory_region(
            region=CommandSenderMachineVertex.
            DATA_REGIONS.COMMANDS_AT_STOP_PAUSE.value,
            size=end_command_size, label='commands with stop pause times')

        vertex.reserve_provenance_data_region(spec)

    @staticmethod
    def get_timed_commands_bytes(timed_commands):
        n_bytes = CommandSenderMachineVertex._N_COMMANDS_SIZE
        n_bytes += (
            (CommandSenderMachineVertex._COMMAND_TIMESTAMP_SIZE +
             CommandSenderMachineVertex._COMMAND_WITH_PAYLOAD_SIZE) *
            len(timed_commands)
        )
        return n_bytes

    @staticmethod
    def get_n_command_bytes(commands):
        n_bytes = CommandSenderMachineVertex._N_COMMANDS_SIZE
        n_bytes += (
            CommandSenderMachineVertex._COMMAND_WITH_PAYLOAD_SIZE *
            len(commands)
        )
        return n_bytes

    @overrides(AbstractHasAssociatedBinary.get_binary_file_name)
    def get_binary_file_name(self):
        """
        Return a string representation of the models binary

        """
        return self.BINARY_FILE_NAME

    @overrides(AbstractHasAssociatedBinary.get_binary_start_type)
    def get_binary_start_type(self):
        return ExecutableStartType.USES_SIMULATION_INTERFACE

    @staticmethod
    def get_number_of_mallocs_used_by_dsg():
        return CommandSenderMachineVertex.TOTAL_REQUIRED_MALLOCS



from six import add_metaclass

from spinn_utilities.abstract_base import AbstractBase
from spinn_utilities.abstract_base import abstractmethod


@add_metaclass(AbstractBase)
class AbstractProvidesKeyToAtomMapping(object):
    """ interface to provide a mapping between routing key partitions and\
        atom ids

    """

    __slots__ = ()

    @abstractmethod
    def routing_key_partition_atom_mapping(self, routing_info, partition):
        """ returns a list of atom to key mapping.

        :param routing_info: the routing info object to consider
        :param partition: the routing partition to handle.
        :return: a iterable of tuples of atom ids to keys.
        """



from spinn_utilities.progress_bar import ProgressBar

# front end common imports
from spinn_front_end_common.interface.provenance \
    import AbstractProvidesProvenanceDataFromMachine
import logging

logger = logging.getLogger(__name__)


class PlacementsProvenanceGatherer(object):
    __slots__ = []

    def __call__(
            self, transceiver, placements, has_ran,
            provenance_data_objects=None):
        """
        :param transceiver: the SpiNNMan interface object
        :param placements: The placements of the vertices
        :param has_ran: token that states that the simulation has ran
        """

        if not has_ran:
            logger.warning("{} skipped as nothing has run "
                           "".format(self.__class__.__name__))
            return list()

        if provenance_data_objects is not None:
            prov_items = provenance_data_objects
        else:
            prov_items = list()

        progress = ProgressBar(
            placements.n_placements, "Getting provenance data")

        # retrieve provenance data from any cores that provide data
        for placement in progress.over(placements.placements):
            if isinstance(placement.vertex,
                          AbstractProvidesProvenanceDataFromMachine):
                # get data
                prov_items.extend(
                    placement.vertex.get_provenance_data_from_machine(
                        transceiver, placement))

        return prov_items



from spinn_utilities.progress_bar import ProgressBar

from spinnman.constants import ROUTER_REGISTER_REGISTERS
from spinnman.model import DiagnosticFilter
from spinnman.model.enums \
    import DiagnosticFilterDefaultRoutingStatus, DiagnosticFilterPacketType
from spinnman.model.enums import DiagnosticFilterSource


class RoutingSetup(object):
    __slots__ = []

    def __call__(self, router_tables, app_id, transceiver, machine):
        routing_tables = list(router_tables.routing_tables)
        progress = ProgressBar(routing_tables, "Preparing Routing Tables")

        # Clear the routing table for each router that needs to be set up
        # and set up the diagnostics
        for router_table in progress.over(routing_tables):
            if not machine.get_chip_at(router_table.x, router_table.y).virtual:
                transceiver.clear_multicast_routes(
                    router_table.x, router_table.y)
                transceiver.clear_router_diagnostic_counters(
                    router_table.x, router_table.y)

                # set the router diagnostic for user 3 to catch local default
                # routed packets. This can only occur when the source router
                # has no router entry, and therefore should be detected a bad
                # dropped packet.
                self._set_router_diagnostic_filters(
                    router_table.x, router_table.y, transceiver)

    @staticmethod
    def _set_router_diagnostic_filters(x, y, transceiver):
        transceiver.set_router_diagnostic_filter(
            x, y, ROUTER_REGISTER_REGISTERS.USER_3.value,
            DiagnosticFilter(
                enable_interrupt_on_counter_event=False,
                match_emergency_routing_status_to_incoming_packet=False,
                destinations=[],
                sources=[DiagnosticFilterSource.LOCAL],
                payload_statuses=[],
                default_routing_statuses=[
                    DiagnosticFilterDefaultRoutingStatus.DEFAULT_ROUTED],
                emergency_routing_statuses=[],
                packet_types=[DiagnosticFilterPacketType.MULTICAST]))

        transceiver.set_router_diagnostic_filter(
            x, y, ROUTER_REGISTER_REGISTERS.USER_2.value,
            DiagnosticFilter(
                enable_interrupt_on_counter_event=False,
                match_emergency_routing_status_to_incoming_packet=False,
                destinations=[],
                sources=[DiagnosticFilterSource.NON_LOCAL],
                payload_statuses=[],
                default_routing_statuses=[
                    DiagnosticFilterDefaultRoutingStatus.DEFAULT_ROUTED],
                emergency_routing_statuses=[],
                packet_types=[DiagnosticFilterPacketType.MULTICAST]))



class ProvenanceDataItem(object):
    """ Container for provenance data
    """

    def __init__(self, names, value, report=False, message=None):
        """

        :param names: A list of strings representing the hierarchy of naming\
                    of this item
        :param value: The value of the item
        :param report: True if the item should be reported to the user
        :param message: The message to send to the end user if report is True
        """
        self._names = names
        self._value = value
        self._report = report
        self._message = message

    @property
    def message(self):
        """ The message to report to the end user, or None if no message
        """
        return self._message

    @property
    def report(self):
        """ True if this provenance data entry needs reporting to the end user
        """
        return self._report

    @property
    def names(self):
        """ The hierarchy of names of this bit of provenance data
        """
        return self._names

    @property
    def value(self):
        """ The value of the item
        """
        return self._value

    def __repr__(self):
        return "{}:{}:{}:{}".format(
            self._names, self._value, self._report, self._message)

    def __str__(self):
        if self._report:
            return self._message
        return "{}: {}".format(self._names, self._value)
